/*
 * fixsch - replace non-verilog net names, etc.
 *
 * Copyright (C) 2007, 2017  Adrian Wise
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See http://www.series16.adrianwise.co.uk
 *   mailto:apw@adrianwise.co.uk
 *
 */
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <list>
#include <map>
#include <cctype>

using namespace std;

static string &tidy_name(string &name);
static string &fix_name(string &name);
static string &new_fix_name(const string &name);

enum PARSE_STATE {
  PARSE_NONE,
  PARSE_SIGNAL_LIST,
  PARSE_MODULE_BODY,
  PARSE_INSTANCE
};

class Wire;
class Instance;
class Module;

class Pin {
public:
  Pin(Instance *instance, const string &pin_name, Wire *wire);
  bool is_output() const {return output;};
  const string &get_pin_name() const {return pin_name;};
  Instance *get_instance() const {return instance;};
  void write(ostream &os, bool first) const;
  void set_wire(Wire *new_wire){wire = new_wire;};
  Wire *get_wire() const {return wire;};
private:
  Instance *instance;
  const string pin_name;
  Wire *wire;
  const bool output;

  static bool is_output_name(const string pin_name, const Instance *instance);
};

ostream &operator<<(ostream &os, const Pin&) {
  os << "pin!";
  return os;
}

class Instance {
public:
  Instance(Module *module,
           const string &instance_name,
           const string &module_name);
  ~Instance();
  void new_pin(const string pin_name, const string wire_name);
  const string &get_instance_name() const {return instance_name;};
  const string &get_module_name() const {return module_name;};
  void complete();
  void write (ostream &os) const;
  void name_signal(const string &pin_name, const string &wire_name);  
  void tie_pin(bool supply, const string &pin_name);
  Pin *find_pin(const string &pin_name);
  Wire *connect_pin(const string &pin_name, Wire &wire);  
  unsigned int output_pins() const {return opins;}
  
private:
  Module *module;
  const string instance_name;
  const string module_name;
  map<const string, Pin *> pins;
  unsigned int opins;
};

class Wire {
public:
  enum IO {
    IO_NONE,
    IO_INPUT,
    IO_OUTPUT,
    IO_INOUT
  };
  Wire(Module *module, const string &name, IO io = IO_NONE);
  bool is_port() const {return io != IO_NONE;};
  IO get_io() const {return io;};
  void connect(Pin *pin);
  void rename(const string &name);
  bool decide_name();
  void check_drivers();
  const string &get_name() const {return name;}
  const string &final_name() const;
  list <Pin *> &get_pins() {return pins;};
  void append_pins(list <Pin *> &additional_pins);
  void tie(bool supply);
  void assign(const Wire &from_wire);
  void delete_pin(Pin &pin);
  void add_pin(Pin &pin);

  enum ACTION {
    ACTION_MODULE,
    ACTION_INPUT,
    ACTION_OUTPUT,
    ACTION_INOUT,
    ACTION_WIRE,
    ACTION_ASSIGN
  };
  void write(ostream &os, ACTION action, bool &first) const;
private:
  Module *module;
  const string name;
  const IO io;
  list<Pin *> pins;
  bool renamed;
  string new_name;
};

class Module {
public:
  Module(const string &name);
  void new_wire(const string &name, Wire::IO io);
  Instance *new_instance(const string &module_name,
                         const string &instance_name);
  Wire *find_wire(const string &name) const;
  Instance *find_instance(const string &name) const;
  void name_wires();
  void write(ostream &os) const;
  void prefered_name(const string &name);
  void name_signal(const string &instance_name,
                   const string &pin_name,
                   const string &wire_name);
  const string &select_prefered_driver(const vector<Pin *> &drivers);

  void checks();
  void tie_wire(bool supply, const string &wire_name);
  void assign_wire(const string &wire_name, const string &from_name);
  void tie_pin(bool supply, const string &instance_name,
                   const string &pin_name);
  void add_tie(const string &wire_name, bool supply);
  void add_assign(const string &wire_name, const string &from_name);
  const string &next_tie_name(bool supply);
  bool is_tied(const string &wire_name) const;
  void connect_pin(const string &instance_name,
                   const string &pin_name,
                   const string &signal_name);
private:

  // Hmm - not sure about this
  class AssistItem {
  public:
    enum ASSIST {
      ASSIST_NAME_SIGNAL
    };
    AssistItem(const string &instance_name,
               const string &pin_name,
               const string &wire_name,
               enum ASSIST action = ASSIST_NAME_SIGNAL);
  private:
    const enum ASSIST action;
    const string &instance_name;
    const string &pin_name;
    const string &wire_name;
    //
    bool used;
  };

  const string name;
  map<const string, Wire *> wires;
  map<const string, Instance *> instances;
  map<const string, bool> prefered_names;
  list<AssistItem *> assist_items;
  map<const string, int> ties;
  map<const string, string> assigns;

  void write_wires(ostream &os, Wire::ACTION action) const;
  void write_ties(ostream &os) const;
  void write_assigns(ostream &os) const;
  void check_unused_prefered();

  int num_tie_low;
  int num_tie_high;
};

bool Pin::is_output_name(const string pin_name, const Instance *instance)
{
  bool output = false;
  int n = 0;
  int j = 0;
  string::size_type loc;

  /*
   * This is a kludge
   *
   * The "node" of a gate is an input to which diodes can be attached
   * to expand the number of inputs.
   * The set of diodes connected to that node input are themselves
   * usually drawn as an AND gate with a special node output.
   * Unfortunately I decided to call this output "nd" - the same name
   * as the node *input* to the expandable NAND gates. This was a
   * poor choice since it stops the pin name alone denoting whether
   * a pin named "nd" is an input or an output. Look at the name of
   * the gate driving the pin to decide.
   */
  const bool is_and_gate = (instance->get_module_name().find("ccc_and", 0) == 0);

  output = ((pin_name == "op")  || // Normal gate output
            ((pin_name == "nd") && is_and_gate)  || // Node output
            (pin_name == "pu")  || // Pull-up output
            // Outputs from flip-flops
            (pin_name == "ops") ||
            (pin_name == "opr") ||
            // Outputs from CC-091
            (pin_name == "opa") ||
            (pin_name == "ona") ||
            (pin_name == "opb") ||
            (pin_name == "onb") ||
            (pin_name == "opc") ||
            (pin_name == "onc") ||
            (pin_name == "opd") ||
            (pin_name == "ond") ||
            (pin_name == "ope") ||
            (pin_name == "one") ||
            (pin_name == "opf") ||
            (pin_name == "onf") ||
            (pin_name == "opg") ||
            (pin_name == "ong") ||
            (pin_name == "oph") ||
            (pin_name == "onh") ||
            // Output from CC-044
            (pin_name == "opx") ||
            (pin_name == "onx")
            );

  if (!output) {

    j = pin_name.size();
    
    if ((pin_name.size() > 2) &&
        ((pin_name.substr(0,2) == "op") ||
         (pin_name.substr(0,2) == "oq") ||
         (pin_name.substr(0,2) == "on") ||
         (pin_name.substr(0,2) == "pu") )) {
      n = 2;
    } else if ((pin_name.size() > 3) &&
               (pin_name.substr(0,3) == "tap")) {
      n = 3;
      if (j > 4) {
        // e.g. tap1a
        if (pin_name[j-1] == 'a') {
          --j;
        }
      }
    }
  
    if (n > 0) {
      int i;
      bool numeric=true;
      for (i=n; ((i<j) && numeric); i++) {
        if (!isdigit(pin_name[i])) {
          numeric = false;
        }
      }
      if (numeric) {
        output = true;
      }
    }
  }

  return output;
}

Pin::Pin(Instance *instance, const string &pin_name, Wire *wire)
  : instance(instance),
    pin_name(pin_name),
    wire(wire),
    output(is_output_name(pin_name, instance))
{
  wire->connect(this);
}

void Pin::write(ostream &os, bool first) const
{
  string wname = wire->final_name();

  if (!first)
    os << " ," << endl;
  os << "." << pin_name << " ( "
     << fix_name(wname) << " )"; 
}

Instance::Instance(Module *module,
                   const string &instance_name,
                   const string &module_name)
  : module(module),
    instance_name(instance_name),
    module_name(module_name),
    opins(0)
{
}

Instance::~Instance()
{
}

void Instance::write(ostream &os) const
{
  string iname = instance_name;

  os << module_name << " " << "u_" << fix_name(iname)
     << " (" << endl;
  
  //list<Pin *>::const_iterator i;
  map<const string, Pin *>::const_iterator i;
  bool first = true;
  for(i = pins.begin(); i != pins.end(); i++) {
    i->second->write(os, first);
    first = false;
  }

  os << endl << ");" << endl << endl;
}

Wire::Wire(Module *module, const string &name, IO io)
  : module(module),
    name(name),
    io(io),
    renamed(false)
{
}

void Wire::connect(Pin *pin)
{
  pins.push_back(pin);
}

const string &Wire::final_name() const
{
  if (renamed)
    return new_name;
  else
    return name;
}

void Wire::append_pins(list <Pin *> &additional_pins)
{
  list <Pin *>::iterator i;
  for (i=additional_pins.begin(); i!=additional_pins.end(); i++) {
    (*i)->set_wire(this);
  }
  pins.splice(--pins.end(), additional_pins);
}

void Wire::delete_pin(Pin &pin)
{
  list <Pin *>::iterator i = pins.begin();

  while (i!=pins.end()) {
    if ((*i) == &pin) {
      break;
    }
    i++;
  }
  
  if (i == pins.end()) {
    cerr << "Failed to find pin " << pin.get_pin_name() << endl;
    exit(1);
  }
  pins.erase(i);
}

void Wire::add_pin(Pin &pin)
{
  pin.set_wire(this);
  pins.push_back(&pin);
}

void Wire::rename(const string &name)
{
  new_name = name;
  renamed = true;
}

void Wire::tie(bool supply)
{
  module->add_tie(final_name(), supply);
}

void Wire::assign(const Wire &from_wire)
{
  module->add_assign(final_name(), from_wire.final_name());
}

Module::Module(const string &name)
  : name(name),
    num_tie_low(0),
    num_tie_high(0)
{
}

void Module::add_tie(const string &wire_name, bool supply)
{
  map<const string, int>::iterator i;
  i = ties.find(wire_name);
  if (i == ties.end())
    ties[wire_name] = supply;
  else {
    cerr << "Already have a signal named <"
         << wire_name << "> in the ties list" << endl;
    exit(1);
  }
}

void Module::add_assign(const string &wire_name, const string &from_name)
{
  map<const string, string>::iterator i;
  i = assigns.find(wire_name);
  if (i == assigns.end())
    assigns[wire_name] = from_name;
  else {
    cerr << "Already have a signal named <"
         << wire_name << "> in the assigns list" << endl;
    exit(1);
  }
}

const string &Module::next_tie_name(bool supply)
{
  ostringstream s;
  string *str;
  int i;
  
  i = (supply) ? (++num_tie_low) : (++num_tie_high);
  s << ((supply) ? "tiehi" : "tielo") << i;
  str = new string(s.str());
  return *str;
}

bool Module::is_tied(const string &wire_name) const
{
  map<const string, int>::const_iterator i;
  i = ties.find(wire_name);
  if (i == ties.end()) {
    map<const string, string>::const_iterator i;
    i = assigns.find(wire_name);
    return (i != assigns.end());
  }
  return true;  
}

void Module::new_wire(const string &name, Wire::IO io)
{
  map<const string, Wire *>::iterator i;
  i = wires.find(name);
  if (i == wires.end()) {
    wires[name] = new Wire(this, name, io);
  } else {
    // It is OK to have an input/output/inout
    // declaration followed by a wire of the
    // name. And we also ignore a second definiton
    // of a port, if the directio matches.

    if (!(i->second->is_port() &&
          ((io == Wire::IO_NONE) || (io == i->second->get_io())))) {
      cerr << "Wire <" << name << "> already declared" << endl;
      exit(1);
    }
  }
}

Wire *Module::find_wire(const string &name) const
{
  map<const string, Wire *>::const_iterator i;
  i = wires.find(name);
  if (i == wires.end())
    return 0;
  else
    return i->second;
}

Instance *Module::find_instance(const string &name) const
{
  map<const string, Instance *>::const_iterator i;
  i = instances.find(name);
  if (i == instances.end())
    return 0;
  else
    return i->second;
}

void Module::tie_wire(bool supply, const string &wire_name)
{
  Wire *wire = find_wire(wire_name);
  if (wire)
    wire->tie(supply);
  else {
    cerr << "tie: Could not find wire named <"
         << wire_name << ">" << endl;
    exit(1);
  }
}

void Module::assign_wire(const string &wire_name, const string &from_name)
{
  Wire *wire = find_wire(wire_name);
  if (wire) {
    Wire *from_wire = find_wire(from_name);
    if (from_wire) {
      wire->assign(*from_wire);
    } else {
      cerr << "assign: Could not find \"from\" wire named <"
           << from_name << ">" << endl;
    exit(1);
  }
  } else {
    cerr << "assign: Could not find \"to\" wire named <"
         << wire_name << ">" << endl;
    exit(1);
  }
}

void Module::tie_pin(bool supply, const string &instance_name,
             const string &pin_name)
{
  Instance *instance = find_instance(instance_name);
  if (instance)
    instance->tie_pin(supply, pin_name);
  else {
    cerr << "tie: Could not find instance named <"
         << instance_name << ">" << endl;
    exit(1);
  }
}

Instance *Module::new_instance(const string &module_name,
                               const string &instance_name)
{
  Instance *r = 0;
  map<const string, Instance *>::iterator i;
  i = instances.find(instance_name);
  if (i == instances.end()) {
    r = new Instance(this, instance_name, module_name);
    instances[instance_name] = r;
  } else {
    cerr << "Instance <" << instance_name << "> already defined" << endl;
    exit(1);
  }
  return r;
}

void Module::connect_pin(const string &instance_name,
                         const string &pin_name,
                         const string &signal_name)
{
  Instance *instance = find_instance(instance_name);
  if (instance) {
    Wire *w = find_wire(signal_name);
    
    if (!w) {
      bool is_supply = false;
      bool supply = false;
      
      if (signal_name == "VDD") {
        is_supply = true;
        supply = true;
      } else if (signal_name == "GND") {
        is_supply = true;
        supply = false;
      }

      if (is_supply) {
        w = new Wire(this, next_tie_name(supply), Wire::IO_NONE);
        w->tie(supply);    
      }
    }
    
    if (w) {
      Wire *current_wire = instance->connect_pin(pin_name, *w);

      if (current_wire &&
          current_wire->get_pins().empty() &&
          (current_wire->get_io() == Wire::IO_NONE)) {
        wires.erase(current_wire->get_name());
      }
      
    } else {
      cerr << "connect: Could not find signal named <"
           << signal_name << ">" << endl;
      exit(1);
    }
  } else {
    cerr << "connect: Could not find instance named <"
         << instance_name << ">" << endl;
    exit(1);
  }
}

void Instance::new_pin(const string pin_name, const string wire_name)
{
  string p = pin_name;
  map<const string, Pin *>::iterator i;

  //cerr << "Debug: new_pin(" << pin_name << ", " << wire_name << ")" << endl;
  
  if (p[0] == '.') {
    p.erase(0,1);
  }

  Wire *w = module->find_wire(wire_name);
  if (w) {
    i = pins.find(p);
    if (i == pins.end()) {
      pins[p] = new Pin(this, p, w);
    } else {
      cerr << "Instance <" << instance_name
           << "> already has a pin named <"
           << p << ">" << endl;
      exit(1);
    }
  } else {
    cerr << "Could not find wire <" << wire_name << ">" << endl;
    exit(1);
  }        

}

void Instance::complete()
{
  map<const string, Pin *>::iterator i;
  int num;
  //cerr << "complete()" << endl;

  // Count the number of output pins
  opins = 0;
  i = pins.begin();
  do {
    if (i->second->is_output()) {
      ++opins;
    }
    ++i;
  } while (i != pins.end());
  
  if (module_name.find("ccc_nand", 0) == 0) {
    num = module_name[8] - '0' + 1;
    if ((module_name.size()>9) && (module_name[9]=='n'))
      num++; // Node in/out
    if (pins.size() != num) {
      cerr << "Instance <" << instance_name << "> of <" <<
        module_name << "> has " << pins.size() <<
        " pins, should have " << num << endl;
      if (pins.size() != num-1) {
        cerr << "Missing more than one pin" << endl;
        exit(1);
      }
    }
  }
}

Pin *Instance::find_pin(const string &pin_name)
{
  map<const string, Pin *>::iterator i;
  i = pins.find(pin_name);
  if (i == pins.end())
    return 0;
  else
    return i->second;
}

void Instance::name_signal(const string &pin_name, const string &wire_name)
{
  Pin *pin = find_pin(pin_name);

  if (pin)
    pin->get_wire()->rename(wire_name);
  else {
    cerr << "Failed to find pin named <" << pin_name << ">" << endl;
    exit(1);
  }
}  

void Instance::tie_pin(bool supply, const string &pin_name)
{
  Pin *pin = find_pin(pin_name);
  Wire *wire;

  if (pin) {
    wire = pin->get_wire();
    //cerr << "Renaming <" << wire->final_name() << ">" << endl;
    wire->rename(module->next_tie_name(supply));
    wire->tie(supply);    
  } else {
    cerr << "Failed to find pin named <" << pin_name << ">" << endl;
    exit(1);
  }
}

Wire *Instance::connect_pin(const string &pin_name, Wire &wire)
{
  Pin *pin = find_pin(pin_name);
  Wire *current_wire = NULL;
  
  if (pin) {
    current_wire = pin->get_wire();
    
    current_wire->delete_pin(*pin);
    wire.add_pin(*pin);
  } else {
    cerr << "Failed to find pin named <" << pin_name << ">" << endl;
    exit(1);
  }
  return current_wire;
}

void Module::name_wires()
{
  map<const string, Wire *>::iterator i, j;
  bool renamed;
  list<map<const string, Wire *>::iterator> to_be_erased;
  list<map<const string, Wire *>::iterator>::iterator k;

  for (i = wires.begin(); i != wires.end(); i++) {
    renamed = i->second->decide_name();

    if (renamed) {
      j = wires.find(i->second->final_name());
      
      if ((j != wires.end()) &&
          (j->second != i->second)) {
        //cout << "Debug: renamed wire <" <<
        //  i->second->final_name() << "> already exists" << endl;
        j->second->append_pins(i->second->get_pins());
        to_be_erased.push_back(i);
      }
    }
  }

  for (k = to_be_erased.begin(); k != to_be_erased.end(); k++) {
    delete (*k)->second;
    wires.erase(*k);
  }

  for (i = wires.begin(); i != wires.end(); i++)
    i->second->check_drivers();

}

Module::AssistItem::AssistItem(const string &instance_name,
                               const string &pin_name,
                               const string &wire_name,
                               enum ASSIST action)
  : instance_name(instance_name),
    pin_name(pin_name),
    wire_name(wire_name),
    action(action),
    used(false)
{
}

bool Wire::decide_name()
{
  list<Pin *>::iterator i;
  vector<Pin *> drivers;
  //vector<Pin *>::iterator j;
  string::size_type loc;
  bool unnamed = false;

  //cout << "Debug: decide_name() Working on net <"
  // << name << ">" << endl;

  if (!renamed) {

    loc = name.find("unnamed_net", 0);
    if (loc == 0)
      unnamed = true;

    i=pins.begin();
    while (i!=pins.end()){
      if ((*i)->is_output())
        drivers.push_back(*i);
      i++;
    }
    
    //cout << "Debug: Wire <" << name << "> has "
    // << drivers.size() << " drivers" << endl;
    if (unnamed) {
      if (drivers.size() == 0) {
        // Ought to check primary inputs!
        //cerr << "Warning: <" << name << "> has no drivers (yet)" << endl;    
      } else if (drivers.size() == 1) {
        // Propogate instance name to signal name
        Instance *inst = drivers[0]->get_instance();
        renamed = true;

        new_name = inst->get_instance_name();
        
        if (inst->output_pins() > 1) {
          // Deal with cases of multiple outputs
          const char ch = *new_name.rbegin();
          if ((ch != '-') && (ch != '+')) {
            new_name += "_";
          }
          new_name += drivers[0]->get_pin_name();
        }

        //cout << "Debug: Renamed Wire <" << name << "> to <" << new_name << ">" << endl;
      } else {
        //cout << "Debug: Wire <" << name << "> has MULTIPLE drivers" << endl;

        new_name = module->select_prefered_driver(drivers);
        renamed = true;
      }
    }
  }
  return renamed;
}

void Wire::check_drivers()
{
  list<Pin *>::iterator i;
  vector<Pin *> drivers;

  i=pins.begin();
  while (i!=pins.end()){
    if ((*i)->is_output())
      drivers.push_back(*i);
    i++;
  }

  if ((drivers.size() == 0) &&
      (io != IO_INPUT) &&
      (io != IO_INOUT) &&
      (!module->is_tied(final_name()))) {
    cout << "Warning: <" << final_name() <<
      "> has no drivers";
    if (pins.size() == 0)
      cout << " and no loads" << endl;
    else {
      cout << ", connects to:" << endl;
      i=pins.begin();
      while (i!=pins.end()){
        cout << "         Pin " << (*i)->get_pin_name()
             << " of instance "
             << (*i)->get_instance()->get_instance_name() << endl;
        i++;
      }
    }
  }
}

void Wire::write(ostream &os, ACTION action, bool &first) const
{
  string s = final_name();
  s = fix_name(s);

  switch (action) {
  case ACTION_MODULE:
    if (is_port()) {
      if (!first)
        os << " ," << endl;
      os << s;
      first = false;
    }
    break;
  case ACTION_INPUT:
    if (io == IO_INPUT)
      os << "input " << s << " ;" << endl;
    break;
  case ACTION_OUTPUT:
    if (io == IO_OUTPUT)
      os << "output " << s << " ;" << endl;
    break;
  case ACTION_INOUT:
    if (io == IO_INOUT)
      os << "inout " << s << " ;" << endl;
    break;
  case ACTION_WIRE:
    if (! is_port())
      os << "wire " << s << " ;" << endl;
    break;
  case ACTION_ASSIGN:
    cerr << "Not implemented" << endl;
    exit(2);
  default:
    cerr << "Unexpected action" << endl;
    exit(2);
  }
}
 
void Module::write_wires(ostream &os, Wire::ACTION action) const
{
  map<const string, Wire *>::const_iterator i;
  bool first = true;

  for (i=wires.begin(); i!=wires.end(); i++)
    i->second->write(os, action, first);
}

void Module::write_ties(ostream &os) const
{
  string wname;
  map<const string, int>::const_iterator i;
  for (i=ties.begin(); i!=ties.end(); i++) {
    wname = i->first;
    os << "assign " << fix_name(wname) << " = "
       << ((i->second) ? "1'b1" : "1'b0") << ";" << endl;
  }
  os << endl;
}

void Module::write_assigns(ostream &os) const
{
  string wname;
  string fname;
  map<const string, string>::const_iterator i;
  for (i=assigns.begin(); i!=assigns.end(); i++) {
    wname = i->first;
    fname = i->second;
    os << "assign " << fix_name(wname) << " = "
       << fix_name(fname) << ";" << endl;
  }
  os << endl;
}

void Module::write(ostream &os) const
{
  os << "module " << name << " (" << endl;
  write_wires(os, Wire::ACTION_MODULE);
  os << endl << " );" << endl << endl;

  write_wires(os, Wire::ACTION_INPUT);
  os << endl;
  write_wires(os, Wire::ACTION_INOUT);
  os << endl;
  write_wires(os, Wire::ACTION_OUTPUT);
  os << endl;
  write_wires(os, Wire::ACTION_WIRE);
  os << endl;

  write_ties(os);
  write_assigns(os);

  map<const string, Instance *>::const_iterator i;
  //cout << "size = " << instances.size() << endl; 
  for (i=instances.begin(); i!=instances.end(); i++) {
    i->second->write(os);
  }

  os << "endmodule" << endl;
}

void Module::prefered_name(const string &name)
{
  map<const string, bool>::iterator i = prefered_names.find(name);
  if (i == prefered_names.end()) {
    prefered_names.insert(make_pair(name, *new bool));
  } else {
    cerr << "Name <" << name
         << "> is already in list of prefered names" << endl;
    exit(1);
  }
}

void Module::name_signal(const string &instance_name,
                         const string &pin_name,
                         const string &wire_name)
{
//   AssistItem *item = new AssistItem(instance_name,
//                                  pin_name,
//                                  wire_name);
//   assist_items.push_back(item);
  
  map<const string, Instance *>::iterator i;

  i = instances.find(instance_name);
  if (i == instances.end()) {
    cerr << "Can't find assist instance name <"
         << instance_name << ">" << endl;
    exit(1);
  }

  i->second->name_signal(pin_name, wire_name);  
}

const string &Module::select_prefered_driver(const vector<Pin *> &drivers)
{
  vector<Pin *>::const_iterator i;
  const Wire *wire = 0;
  const Wire *wp;

  int count_ff_names = 0;
  const string *ff_name;

  int count_prefered_names = 0;
  map<const string, bool>::iterator ipn, tpn;
  
  const Instance *inst;

  //
  // Run through the drivers, check that they're consistently
  // all connected to the same wire.
  // See if any of the driving instances look like flip-flop
  // names, or any are in the prefered names list
  //
  for (i = drivers.begin(); i != drivers.end(); i++) {

    wp = (*i)->get_wire();
    if (wire && (wire != wp)) {
      cerr << "Not all these drivers connect to the same wire" << endl;
      exit(2);
    }
    wire = wp;

    inst = (*i)->get_instance();
    const string &inst_name = inst->get_instance_name();

    string::size_type loc = inst_name.find("FF", 3);
    if ((loc == 3) || (loc == 4)) {
      count_ff_names++;
      ff_name = &inst_name;
    }

    tpn = prefered_names.find(inst_name);
    if (tpn != prefered_names.end()) {
      count_prefered_names++;
      ipn = tpn;
    }
  }

  if (count_prefered_names > 1) {
    cerr << "Error: Multiple prefered names match for wire <"
         << ((wire) ? wire->final_name() : "(unknown)")
         << ">" << endl;
    exit(1);
  } else if (count_prefered_names == 1) {
    ipn->second = true;
    return ipn->first;
  } else if (count_ff_names > 1) {
    cerr << "Error: Multiple flip-flop names for wire <"
         << ((wire) ? wire->final_name() : "(unknown)")
         << ">" << endl;
    exit(1);    
  } else if (count_ff_names == 1) {
    return *ff_name;
  } else {
    cerr << "Warning: <" << wire->final_name() << "> has " 
         << drivers.size()
         << " drivers, but none are listed as prefered or are flip-flops"
         << endl;
    cerr << "       : Candidate names are:" << endl;
    for (i = drivers.begin(); i != drivers.end(); i++) {
      inst = (*i)->get_instance();
      const string &inst_name = inst->get_instance_name();
      cerr << "       : <" << inst_name << ">";
      if (i == drivers.begin()) {
        ff_name = &inst_name;
        cerr << " (selected)" << endl;
      } else
        cerr << endl;
    }
    return *ff_name;
  }
}

void Module::checks()
{
  check_unused_prefered();
}

void Module::check_unused_prefered()
{
  map<const string, bool>::iterator i;
  for(i=prefered_names.begin(); i!=prefered_names.end(); i++) {
    if (! i->second) {
      cerr << "Warning: Prefered name <"
           << i->first << "> remains unused" << endl;
    }
  }

}

static string *readstr(istream &is)
{
  char c;

  // Discard leading white space
  do {
    is.get(c);
  } while (is.good() &&
           isspace(c) && (c!='\n'));

  if (is.good()) {
    string *str = new string;

    if (c == '\n') {
      return str;
    } else if (c == '"') {
      do {
        (*str) += c;
        is.get(c);
      } while (is.good() && (c != '"'));
      (*str) += c;
    } else {
      do {
        (*str) += c;
        is.get(c);
      } while(is.good() && (!isspace(c)));
      is.unget();
    }
    return str;
  } else {
    return 0;
  }
}

static vector<string *> *readline(istream &is)
{
  string *str = readstr(is);

  if (!str)
    return 0; // EOF

  vector<string *> *tmp = new vector<string *>;

  while ((str) && (str->length()>0)) {
    tmp->push_back(str);
    str=readstr(is);
  }
  
  if (str)
    delete str; // The newline string

  return tmp;
}

static void printline(ostream &os, vector<string *> *line)
{
  int i=0;
  if (line) {
    while (i < line->size())
      os << (*(*line)[i++]) << ' ';
    os << endl;
  } else {
    os << "EOF" << endl;
  }
}

static void deleteline(vector<string *> *line)
{
  vector<string *>::iterator i;
  for (i=line->begin(); i!=line->end(); i++)
    delete *i;
  delete line;
}

static string &tidy_name(string &name)
{
  if (name[0] == '\\')
    name.erase(0,1);
  return name;
}

static string &fix_name(string &name)
{
  string fixed_name = name;
  string::size_type loc;

  if ((fixed_name.size() > 0) &&
      (isdigit(fixed_name[0]))) {
    cerr << "Warning: name <" << name
         << "> starts with a digit." << endl;
    fixed_name.insert(0, "_");
  }

  if ((loc = fixed_name.find('-', 0)) != string::npos) {
    fixed_name[loc] = '_';
    fixed_name.insert(loc+1, "n");
    if (loc+2 != fixed_name.size())
      fixed_name.insert(loc+2, "_");
  }
  if ((loc = fixed_name.find('+', 0)) != string::npos) {
    fixed_name[loc] = '_';
    fixed_name.insert(loc+1, "p");
    if (loc+2 != fixed_name.size())
      fixed_name.insert(loc+2, "_");
  }

  name = fixed_name;
  return name;
}

static string &new_fix_name(const string &name)
{
  string new_name(name);
  return fix_name(new_name);
}

static Module *parse_verilog(ifstream &ifs)
{
  vector<string *> *p_line;
  PARSE_STATE parse_state = PARSE_NONE;
  Module *module = 0;
  Instance *instance = 0;

  p_line = readline(ifs);

  while(p_line) {
    //printline(cout, p_line);
    vector<string *> &line = *p_line;

    if (line.size() > 0) {
      if ((*line[0]) == "/*") {
        // Comment - ignore
        // Netlister always produces one line
        // comments so this is trivial
      } else if ((*line[0]) == "module") {
        parse_state = PARSE_SIGNAL_LIST;
        if ((line.size() != 3) ||
            ((*line[2]) != "(")) {
          cerr << "Expected " << (*line[0]) <<
            " <module-name> (" << endl;
          exit(1);
        }
        module = new Module(*line[1]);
      } else if ((*line[0]) == "endmodule") {
        if (parse_state != PARSE_MODULE_BODY) {
          cerr << "Unexpected endmodule" << endl;
          exit(1);
        }
      } else if ((*line[0]) == ");") {
        if (parse_state == PARSE_SIGNAL_LIST) {
          parse_state = PARSE_MODULE_BODY;
        } else if (parse_state == PARSE_INSTANCE) {
          parse_state = PARSE_MODULE_BODY;
          instance->complete();
          instance = 0;
        } else {
          cerr << "Unexpected closing bracket" << endl;
          exit(1);
        }
      } else if (((*line[0]) == "wire") ||
                 ((*line[0]) == "input") ||
                 ((*line[0]) == "output") ||
                 ((*line[0]) == "inout")) {
        Wire::IO io;

        if (parse_state != PARSE_MODULE_BODY) {
          cerr << "Unexpected " << line[0] << endl;
          exit(1);
        }

        if ((*line[0]) == "wire")
          io = Wire::IO_NONE;
        else if ((*line[0]) == "input")
          io = Wire::IO_INPUT;
        else if ((*line[0]) == "output")
          io = Wire::IO_OUTPUT;
        else if ((*line[0]) == "inout")
          io = Wire::IO_INOUT;
        else {
          cerr << "How did that happen?" << endl;
          exit(2);
        }

        if ((line.size() != 3) ||
            ((*line[2]) != ";")) {
          printline(cerr, &line);
          cerr << "Expected " << (*line[0]) <<
            " <signal-name> ;" << endl;
          exit(1);
        }

        module->new_wire(tidy_name(*line[1]), io);

      } else {

        if (parse_state == PARSE_SIGNAL_LIST) {
          // Do nothing
        } else if (parse_state == PARSE_MODULE_BODY) {

          if ((line.size() != 3) ||
              ((*line[2]) != "(")) {
            printline(cerr, &line);
            cerr << "Expected <module-name> <instance-name> (" << endl;
            exit(1);
          }

          //instance = module->new_instance(line[0], line[1]);
          instance = module->new_instance((*line[0]), tidy_name(*line[1]));
          parse_state = PARSE_INSTANCE;

        } else if (parse_state == PARSE_INSTANCE) {

          if ((line.size() != 4) ||
              ((*line[0])[0] != '.') ||
              ((*line[1]) != "(") ||
              (((*line[3]) != ")") && ((*line[3]) != "),"))) {
            printline(cerr, &line);
            cerr << "Expected .<port-name> ( <signal-name> )[,]" << endl;
            exit(1);
          }

          instance->new_pin((*line[0]), tidy_name(*line[2]));

        } else {
          cerr << "Unexpected parse state" << endl;
          exit(2);
        }

      }
    }

    deleteline(p_line);
    p_line = readline(ifs);
  }
  return module;
}

void parse_patchup(Module *module, ifstream &ifs)
{
  vector<string *> *p_line;
 
  p_line = readline(ifs);

  while(p_line) {
    vector<string *> &line = *p_line;

    if (line.size() > 0) {
      if ((*line[0]) == "prefer") {
        if (line.size() == 2) {
          module->prefered_name(*line[1]);
        } else {
          cerr << "Expected one word following <"
               << (*line[0]) << "> got " << line.size()-1 << endl;
          exit(1);
        }
      } else if ((*line[0]) == "name") {
        if (line.size() == 4) {
          module->name_signal(*line[1], *line[2], *line[3]);
        } else {
          cerr << "Expected three words following <"
               << (*line[0]) << "> got " << line.size()-1 << endl;
          exit(1);
        }
      } else if ((*line[0]) == "tiehi") {
        if (line.size() == 2) {
          module->tie_wire(1, *line[1]);
        } else if (line.size() == 3) {
          module->tie_pin(1, *line[1], *line[2]);
        } else {
          cerr << "Expected one or two words following <"
               << (*line[0]) << "> got " << line.size()-1 << endl;
          exit(1);
        }
      } else if ((*line[0]) == "tielo") {
        if (line.size() == 2) {
          module->tie_wire(0, *line[1]);
        } else if (line.size() == 3) {
          module->tie_pin(0, *line[1], *line[2]);
        } else {
          cerr << "Expected one or two words following <"
               << (*line[0]) << "> got " << line.size()-1 << endl;
          exit(1);
        }
      } else if ((*line[0]) == "assign") {
        if (line.size() == 3) {
          module->assign_wire(*line[1], *line[2]);
        } else {
          cerr << "Expected two words following <"
               << (*line[0]) << "> got " << line.size()-1 << endl;
          exit(1);
        }
      } else if ((*line[0]) == "connect") { // connect <instance> <pin> <signal>
        if (line.size() == 4) {
          module->connect_pin(*line[1], *line[2], *line[3]);
        } else {
          cerr << "Expected three words following <"
               << (*line[0]) << "> got " << line.size()-1 << endl;
          exit(1);
        }
      } else if ((*line[0]) == "//") {
        // Discard line
      } else {
        cerr << "Unexpected line starting <"
             << (*line[0]) << ">" << endl;
        exit(1);
      }
    }
    p_line = readline(ifs);
  }
}

int main(int argc, char **argv)
{
  string str;
  bool usage = false;
  int i = 1;
  char *input_filename = 0;
  char *output_filename = 0;
  char *patchup_filename = 0;
  ifstream ifs;
  ofstream ofs;
  Module *module;

  while (i < argc) {
    str = argv[i];

    if (str == "-o") {
      if (++i < argc)
        output_filename = argv[i];
      else
        usage = true;
    } else if (str == "-p") {
      if (++i < argc)
        patchup_filename = argv[i];
      else
        usage = true;      
    } else {
      if (i == (argc-1))
        input_filename = argv[i];
      else
        usage = true;
    }
    i++;
  }

  if (!input_filename)
    usage = true;

  if (usage) {
    cerr << "usage :" << argv[0]
         << " [-o <output-filename>] [-p <patchup-filename>] filename" << endl;
    exit(1);
  }

  ifs.open(input_filename);

  if (!ifs.is_open()) {
    cerr << "could not open <" << input_filename
         << "> for reading" << endl;
    exit(1);
  }
  
  module = parse_verilog(ifs);

  ifs.close();

  if (patchup_filename) {
    ifs.open(patchup_filename);

    if (!ifs.is_open()) {
      cerr << "could not open <" << patchup_filename
           << "> for reading" << endl;
      exit(1);
    }
  
    parse_patchup(module, ifs);

    ifs.close();
  }

  module->name_wires();
  module->checks();

  if (output_filename) {
    ofs.open(output_filename);

    if (!ofs.is_open()) {
      cerr << "could not open <" << output_filename
           << "> for writing" << endl;
      exit(1);
    }

    module->write(ofs);

    if (output_filename)
      ofs.close();
  }

  exit(0);
}
