/*
 * makesch - make a skeleton schematic from text file
 *
 * Copyright (C) 2006, 2007, 2017  Adrian Wise
 *
 * This file is part of makesch.
 *
 * makesch is free software: you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See http://www.series16.adrianwise.co.uk
 *   mailto:apw@adrianwise.co.uk
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "components.hh"

#define PAGE_NUM_OFFSET 800
#define SIGNAL_OFFSET 100
#define BRISTLE_LENGTH 200

/****************************************************************
 *
 * Text
 *
 ***************************************************************/

Text::Text(int x, int y, int color, int size, int visibility,
           int show_name_value, int angle, int alignment, const char *str)
  : x(x),
    y(y),
    color(color),
    size(size),
    visibility(visibility),
    show_name_value(show_name_value),
    angle(angle),
    alignment(alignment),
    str(str)
{
};

void Text::output(int x0, int y0, char *attribute)
{
  printf("T %d %d %d %d %d %d %d %d\n%s\n",
         x0+x,
         y0+y,
         color,
         size,
         (attribute) ? visibility : 0,
         show_name_value,
         angle,
         alignment,
         (attribute) ? fix_attribute(attribute) : str);
};

char *Text::fix_attribute(char *attribute)
{
  static char buf[1024];
  strcpy(buf, str);
  int i = 0;
  
  while (buf[i])
    {
      if (buf[i]=='=')
        {
          buf[i+1] = '\0';
          break;
        }
      i++;
    }
  strcat(buf, attribute);
  return buf;
}

/****************************************************************
 *
 * Pin
 *
 ***************************************************************/

Pin::Pin(int x1, int y1, int x2, int y2, int color)
  : x1(x1),
    y1(y1),
    x2(x2),
    y2(y2),
    color(color)
{
}

void Pin::annotate(int xx, int yy, int x, int y, char *signal, char *page,
                   char *post, bool ff)
{
  int x0, y0;
  x0 = xx-x;
  y0 = yy-y;

  int dir = 1; // An output pin on the right

  if (x1 < x)
    dir = -1; // An input pin on the left
  
  int xz, yz;
  
  if ( (x1 * dir) > (x2 * dir) )
    {
      xz = x1;
      yz = y1;
    }
  else
    {
      xz = x2;
      yz = y2;
    }
  
  if (signal)
    {
      printf("N %d %d %d %d %d\n",
             xz + x0,
             yz + y0,
             xz + (BRISTLE_LENGTH * dir) + x0,
             yz + y0,
             4);
      
      printf("{\n");
      printf("T %d %d %d %d %d %d %d %d\nnetname=%s\n",
             xz + ((BRISTLE_LENGTH + SIGNAL_OFFSET) * dir) + x0,
             yz + y0,
             5, 8, 1, 1, 0,
             (dir < 0) ? 7 : 1,
             signal);
      printf("}\n");
      
      if (page)
        {
          printf("T %d %d %d %d %d %d %d %d\n[%s]\n",
                 xz + (( BRISTLE_LENGTH + SIGNAL_OFFSET + PAGE_NUM_OFFSET) *
                       dir) + x0,
                 yz + y0,
                 9, 8, 1, 1, 0,
                 (dir < 0) ? 7 : 1,
                 page);
        }
    }

  if (post)
    {
      int xoffset;

      if (dir < 0)
        xoffset = 50;
      else
        {
          if (ff)
            xoffset = -50;
          else
            xoffset = -100;
        }

      printf("T %d %d %d %d %d %d %d %d\n%s\n",
             xz + xoffset + x0,
             yz + y0 + 100,
             9, 6, 1, 1, 0,
             (dir < 0) ? 7 : 1,
             post);
    }
  
}
/****************************************************************
 *
 * Component
 *
 ***************************************************************/

Component::Component(const char *name, const char *nickname, enum TYPE type, int x, int y,
                     int num_texts, Text texts[],
                     int num_pins, Pin pins[])
  : name(name),
    nickname(nickname),
    type(type),
    x(x),
    y(y),
    num_texts(num_texts),
    num_pins(num_pins)
{
  int i;
  this->texts = new Text *[num_texts + 1];
  for (i=0; i<num_texts; i++)
    this->texts[i] = &texts[i];
  this->texts[i] = 0;
  
  this->texts = new Text *[num_texts + 1];
  for (i=0; i<num_texts; i++)
    this->texts[i] = &texts[i];
  this->texts[i] = 0;
  
  this->pins = new Pin *[num_pins + 1];
  for (i=0; i<num_pins; i++)
    this->pins[i] = &pins[i];
  this->pins[i] = 0;
}

void Component::instance(int xx, int yy, char **attributes)
{
  int x0, y0;
  
  /*
   * Calculate where the origin of this component should
   * be. (xx, yy) is the notional centre of the component
   * and the offset of that centre from the origin is
   * (x, y)
   */
  x0 = xx-x;
  y0 = yy-y;
  
  printf("C %d %d 0 0 0 %s.sym\n",
                                 x0, y0, name);
  printf("{\n");
  
  int i;
  for (i=0; i<num_texts; i++) {
    char *attr = attributes[i];
    if (strcmp(attr, "*")==0) {
      attr = 0;
    }
    texts[i]->output(x0, y0, attr);
  }
  
  printf("}\n");
}

void Component::annotate(int xx, int yy, int pin_no,
                         char *signal, char *page,
                         char *post)
{
  pins[pin_no]->annotate(xx, yy, x, y, signal, page, post,
                         (strncmp(nickname, "ff", 2)==0));
}



/****************************************************************
 *
 * Tables
 *
 ***************************************************************/

/********************** NAND 2 *********************/
Text ccc_nand2_texts[] =
{
  Text(200, 400, 5, 8, 1, 1, 0, 1, "refdes=??????"),
  Text(200, 600, 5, 8, 1, 1, 0, 1, "location=???"),
  Text(200, 200, 5, 8, 1, 1, 0, 1, "pac=???"),
  Text(700, 200, 5, 8, 1, 1, 0, 1, "ref=??")
};

Pin ccc_nand2_pins[] =
{
        Pin(1100,400, 1200,400, 1),
        Pin(100, 400, 0,   400, 1),
        Pin(100, 200, 0,   200, 1)
};
Component ccc_nand2("ccc_nand2", "nand2", Component::GATE, 600, 400,
                    4, ccc_nand2_texts,
                    3, ccc_nand2_pins);

Pin ccc_nand2a_pins[] =
{
        Pin(1100,400, 1200,400, 1),
        Pin(100, 600, 0,   600, 1),
        Pin(100, 400, 0,   400, 1)
};
Component ccc_nand2a("ccc_nand2a", "nand2a", Component::GATE, 600, 400,
                    4, ccc_nand2_texts,
                    3, ccc_nand2a_pins);

Pin ccc_nand2b_pins[] =
{
        Pin(1100,400, 1200,400, 1),
        Pin(100, 600, 0,   600, 1),
        Pin(100, 200, 0,   200, 1)
};
Component ccc_nand2b("ccc_nand2b", "nand2b", Component::GATE, 600, 400,
                    4, ccc_nand2_texts,
                    3, ccc_nand2b_pins);

/********************** NAND 1 *********************/
Pin ccc_nand1_pins[] =
{
        Pin(1100,400, 1200,400, 1),
        Pin(100, 400, 0,   400, 1),
};
Component ccc_nand1("ccc_nand1", "nand1", Component::GATE, 600, 400,
                    4, ccc_nand2_texts, // NB same as nand2!
                    2, ccc_nand1_pins);


/********************** NAND 3 *********************/

Pin ccc_nand3_pins[] =
{
        Pin(1100,400, 1200,400, 1),
        Pin(100, 600, 0,   600, 1),
        Pin(100, 400, 0,   400, 1),
        Pin(100, 200, 0,   200, 1)
};
Component ccc_nand3("ccc_nand3", "nand3", Component::GATE, 600, 400,
                    4, ccc_nand2_texts, // NB same as nand2!
                    4, ccc_nand3_pins);

/********************** NAND 4 *********************/

Pin ccc_nand4_pins[] =
{
        Pin(1100,400, 1200,400, 1),
        Pin(100, 600, 0,   600, 1),
        Pin(100, 400, 0,   400, 1),
        Pin(100, 200, 0,   200, 1),
        Pin(100,   0, 0,     0, 1)
};
Component ccc_nand4("ccc_nand4", "nand4", Component::GATE, 600, 400,
                    4, ccc_nand2_texts, // NB same as nand2!
                    5, ccc_nand4_pins);

/********************** NAND 5 *********************/

Text ccc_nand5_texts[] =
{
        Text(200, 600, 5, 8, 1, 1, 0, 1, "refdes=??????"),
        Text(200, 800, 5, 8, 1, 1, 0, 1, "location=???"),
        Text(200, 400, 5, 8, 1, 1, 0, 1, "pac=???"),
        Text(700, 400, 5, 8, 1, 1, 0, 1, "ref=??")
};
Pin ccc_nand5_pins[] =
{
        Pin(1100,600, 1200,600, 1),
        Pin(100, 800, 0,   800, 1),
        Pin(100, 600, 0,   600, 1),
        Pin(100, 400, 0,   400, 1),
        Pin(100, 200, 0,   200, 1),
        Pin(100,   0, 0,     0, 1)
};
Component ccc_nand5("ccc_nand5", "nand5", Component::GATE, 600, 600,
                    4, ccc_nand5_texts,
                    6, ccc_nand5_pins);

/********************** NAND 6 *********************/

Text ccc_nand6_texts[] =
{
        Text(200, 800, 5, 8, 1, 1, 0, 1, "refdes=??????"),
        Text(200,1000, 5, 8, 1, 1, 0, 1, "location=???"),
        Text(200, 600, 5, 8, 1, 1, 0, 1, "pac=???"),
        Text(700, 600, 5, 8, 1, 1, 0, 1, "ref=??")
};
Pin ccc_nand6_pins[] =
{
        Pin(1100,800, 1200,800, 1),
        Pin(100,1000, 0,  1000, 1),
        Pin(100, 800, 0,   800, 1),
        Pin(100, 600, 0,   600, 1),
        Pin(100, 400, 0,   400, 1),
        Pin(100, 200, 0,   200, 1),
        Pin(100,   0, 0,     0, 1)
};
Component ccc_nand6("ccc_nand6", "nand6", Component::GATE, 600, 800,
                    4, ccc_nand6_texts,
                    7, ccc_nand6_pins);

/********************** NAND 7 *********************/

Text ccc_nand7_texts[] =
{
        Text(200,1000, 5, 8, 1, 1, 0, 1, "refdes=??????"),
        Text(200,1200, 5, 8, 1, 1, 0, 1, "location=???"),
        Text(200, 800, 5, 8, 1, 1, 0, 1, "pac=???"),
        Text(700, 800, 5, 8, 1, 1, 0, 1, "ref=??")
};
Pin ccc_nand7_pins[] =
{
        Pin(1100,1000, 1200,1000, 1),
        Pin(100,1200, 0,  1200, 1),
        Pin(100,1000, 0,  1000, 1),
        Pin(100, 800, 0,   800, 1),
        Pin(100, 600, 0,   600, 1),
        Pin(100, 400, 0,   400, 1),
        Pin(100, 200, 0,   200, 1),
        Pin(100,   0, 0,     0, 1)
};
Component ccc_nand7("ccc_nand7", "nand7", Component::GATE, 600, 1000,
                    4, ccc_nand7_texts,
                    8, ccc_nand7_pins);

/********************** NAND 8 *********************/

Text ccc_nand8_texts[] =
{
        Text(200,1200, 5, 8, 1, 1, 0, 1, "refdes=??????"),
        Text(200,1400, 5, 8, 1, 1, 0, 1, "location=???"),
        Text(200,1000, 5, 8, 1, 1, 0, 1, "pac=???"),
        Text(700,1000, 5, 8, 1, 1, 0, 1, "ref=??")
};
Pin ccc_nand8_pins[] =
{
        Pin(1100,1200, 1200,1200, 1),
        Pin(100,1400, 0,  1400, 1),
        Pin(100,1200, 0,  1200, 1),
        Pin(100,1000, 0,  1000, 1),
        Pin(100, 800, 0,   800, 1),
        Pin(100, 600, 0,   600, 1),
        Pin(100, 400, 0,   400, 1),
        Pin(100, 200, 0,   200, 1),
        Pin(100,   0, 0,     0, 1)
};
Component ccc_nand8("ccc_nand8", "nand8", Component::GATE, 600, 1200,
                    4, ccc_nand8_texts,
                    9, ccc_nand8_pins);

/********************** NAND 8 *********************/

Text ccc_nand9_texts[] =
{
        Text(200,1400, 5, 8, 1, 1, 0, 1, "refdes=??????"),
        Text(200,1600, 5, 8, 1, 1, 0, 1, "location=???"),
        Text(200,1200, 5, 8, 1, 1, 0, 1, "pac=???"),
        Text(700,1200, 5, 8, 1, 1, 0, 1, "ref=??")
};
Pin ccc_nand9_pins[] =
{
        Pin(1100,1400, 1200,1400, 1),
        Pin(100,1600, 0,  1600, 1),
        Pin(100,1400, 0,  1400, 1),
        Pin(100,1200, 0,  1200, 1),
        Pin(100,1000, 0,  1000, 1),
        Pin(100, 800, 0,   800, 1),
        Pin(100, 600, 0,   600, 1),
        Pin(100, 400, 0,   400, 1),
        Pin(100, 200, 0,   200, 1),
        Pin(100,   0, 0,     0, 1)
};
Component ccc_nand9("ccc_nand9", "nand9", Component::GATE, 600, 1400,
                    4, ccc_nand9_texts,
                    10, ccc_nand9_pins);

/********************** NAND 1n *********************/
Text ccc_nand1n_texts[] =
{
  Text(200, 600, 5, 8, 1, 1, 0, 1, "refdes=??????"),
  Text(200, 800, 5, 8, 1, 1, 0, 1, "location=???"),
  Text(200, 400, 5, 8, 1, 1, 0, 1, "pac=???"),
  Text(700, 400, 5, 8, 1, 1, 0, 1, "ref=??")
};

Pin ccc_nand1n_pins[] =
{
  Pin(1100,600, 1200,600, 1),
  Pin(1000,  0, 1100,  0, 1),
  Pin(100, 600, 0,   600, 1)
};

Component ccc_nand1n("ccc_nand1n", "nand1n", Component::GATE, 600, 600,
                     4, ccc_nand1n_texts,
                     3, ccc_nand1n_pins);

/********************** NAND 2n *********************/
Pin ccc_nand2n_pins[] =
{
  Pin(1100,600, 1200,600, 1),
  Pin(1000,  0, 1100,  0, 1),
  Pin(100, 600, 0,   600, 1),
  Pin(100, 400, 0,   400, 1)
};

Component ccc_nand2n("ccc_nand2n", "nand2n", Component::GATE, 600, 600,
                     4, ccc_nand1n_texts,
                     4, ccc_nand2n_pins);

/********************** NAND 3n *********************/
Pin ccc_nand3n_pins[] =
{
  Pin(1100,600, 1200,600, 1),
  Pin(1000,  0, 1100,  0, 1),
  Pin(100, 800, 0,   800, 1),
  Pin(100, 600, 0,   600, 1),
  Pin(100, 400, 0,   400, 1)
};

Component ccc_nand3n("ccc_nand3n", "nand3n", Component::GATE, 600, 600,
                     4, ccc_nand1n_texts,
                     5, ccc_nand3n_pins);

/********************** NAND 4n *********************/
Pin ccc_nand4n_pins[] =
{
  Pin(1100,600, 1200,600, 1),
  Pin(1000,  0, 1100,  0, 1),
  Pin(100, 800, 0,   800, 1),
  Pin(100, 600, 0,   600, 1),
  Pin(100, 400, 0,   400, 1),
  Pin(100, 200, 0,   200, 1)
};

Component ccc_nand4n("ccc_nand4n", "nand4n", Component::GATE, 600, 600,
                     4, ccc_nand1n_texts,
                     6, ccc_nand4n_pins);

/********************** NAND 5n *********************/
Pin ccc_nand5n_pins[] =
{
  Pin(1100,600, 1200,600, 1),
  Pin(1000,  0, 1100,  0, 1),
  Pin(100, 800, 0,   800, 1),
  Pin(100, 600, 0,   600, 1),
  Pin(100, 400, 0,   400, 1),
  Pin(100, 200, 0,   200, 1),
  Pin(100,   0, 0,     0, 1)
};

Component ccc_nand5n("ccc_nand5n", "nand5n", Component::GATE, 600, 600,
                     4, ccc_nand1n_texts,
                     7, ccc_nand5n_pins);

/********************** NAND 6n *********************/
Pin ccc_nand6n_pins[] =
{
        Pin(1100, 800, 1200, 800, 1),
        Pin(1000, 200, 1100, 200, 1),
        Pin(100,1000, 0,  1000, 1),
        Pin(100, 800, 0,   800, 1),
        Pin(100, 600, 0,   600, 1),
        Pin(100, 400, 0,   400, 1),
        Pin(100, 200, 0,   200, 1),
        Pin(100,   0, 0,     0, 1)
};
Component ccc_nand6n("ccc_nand6n", "nand6n", Component::GATE, 600, 800,
                    4, ccc_nand6_texts,
                    8, ccc_nand6n_pins);

/********************** NAND 7n *********************/
Pin ccc_nand7n_pins[] =
{
        Pin(1100,1000, 1200,1000, 1),
        Pin(1000, 400, 1100, 400, 1),
        Pin(100,1200, 0,  1200, 1),
        Pin(100,1000, 0,  1000, 1),
        Pin(100, 800, 0,   800, 1),
        Pin(100, 600, 0,   600, 1),
        Pin(100, 400, 0,   400, 1),
        Pin(100, 200, 0,   200, 1),
        Pin(100,   0, 0,     0, 1)
};
Component ccc_nand7n("ccc_nand7n", "nand7n", Component::GATE, 600, 1000,
                    4, ccc_nand7_texts,
                    9, ccc_nand7n_pins);


/********************** AND 1 *********************/
Text ccc_and1_texts[] =
{
  Text(200, 600, 5, 8, 1, 1, 0, 1, "refdes=??????"),
  Text(200, 800, 5, 8, 1, 1, 0, 1, "location=???"),
  Text(200, 400, 5, 8, 1, 1, 0, 1, "pac=???"),
};
Pin ccc_and1_pins[] =
{
  Pin(1000,  0, 1100,  0, 1),
  Pin(100, 600, 0,   600, 1),
};
Component ccc_and1("ccc_and1", "and1", Component::GATE, 600, 600,
                    3, ccc_and1_texts,
                    2, ccc_and1_pins);

/********************** AND 2 *********************/

Pin ccc_and2_pins[] =
{
  Pin(1000,  0, 1100,  0, 1),
  Pin(100, 600, 0,   600, 1),
  Pin(100, 400, 0,   400, 1)
};
Component ccc_and2("ccc_and2", "and2", Component::GATE, 600, 600,
                    3, ccc_and1_texts,
                    3, ccc_and2_pins);

/********************** AND 2a *********************/
Pin ccc_and2a_pins[] =
{
  Pin(1000,  0, 1100,  0, 1),
  Pin(100, 800, 0,   800, 1),
  Pin(100, 600, 0,   600, 1)
};
Component ccc_and2a("ccc_and2a", "and2a", Component::GATE, 600, 600,
                    3, ccc_and1_texts,
                    3, ccc_and2a_pins);

/********************** AND 3 *********************/

Pin ccc_and3_pins[] =
{
  Pin(1000,  0, 1100,  0, 1),
  Pin(100, 800, 0,   800, 1),
  Pin(100, 600, 0,   600, 1),
  Pin(100, 400, 0,   400, 1),
};
Component ccc_and3("ccc_and3", "and3", Component::GATE, 600, 600,
                    3, ccc_and1_texts,
                    4, ccc_and3_pins);

/********************** AND 4 *********************/

Pin ccc_and4_pins[] =
{
  Pin(1000,  0, 1100,  0, 1),
  Pin(100, 800, 0,   800, 1),
  Pin(100, 600, 0,   600, 1),
  Pin(100, 400, 0,   400, 1),
  Pin(100, 200, 0,   200, 1),
};
Component ccc_and4("ccc_and4", "and4", Component::GATE, 600, 600,
                    3, ccc_and1_texts,
                    5, ccc_and4_pins);

/********************** AND 5 *********************/

Pin ccc_and5_pins[] =
{
  Pin(1000,  0, 1100,  0, 1),
  Pin(100, 800, 0,   800, 1),
  Pin(100, 600, 0,   600, 1),
  Pin(100, 400, 0,   400, 1),
  Pin(100, 200, 0,   200, 1),
  Pin(100,   0, 0,     0, 1)
};
Component ccc_and5("ccc_and5", "and5", Component::GATE, 600, 600,
                    3, ccc_and1_texts,
                    6, ccc_and5_pins);

/********************** AND 9 *********************/

Text ccc_and9_texts[] =
{
  Text(200,1400, 5, 8, 1, 1, 0, 1, "refdes=??????"),
  Text(200,1600, 5, 8, 1, 1, 0, 1, "location=???"),
  Text(200,1200, 5, 8, 1, 1, 0, 1, "pac=???"),
};
Pin ccc_and9_pins[] =
{
  Pin(1000,800, 1100,800, 1),
  Pin(100,1600, 0,  1600, 1),
  Pin(100,1400, 0,  1400, 1),
  Pin(100,1200, 0,  1200, 1),
  Pin(100,1000, 0,  1000, 1),
  Pin(100, 800, 0,   800, 1),
  Pin(100, 600, 0,   600, 1),
  Pin(100, 400, 0,   400, 1),
  Pin(100, 200, 0,   200, 1),
  Pin(100,   0, 0,     0, 1)
};
Component ccc_and9("ccc_and9", "and9", Component::GATE, 600, 1400,
                    3, ccc_and9_texts,
                    10, ccc_and9_pins);

/********************** FF2223  *********************/

Text ccc_ff_texts[] =
{
  Text(500,1900, 5, 8, 1, 1, 0, 4, "refdes=??????"),
  Text(500,2200, 5, 8, 1, 1, 0, 4, "location=???"),
  Text(200, 200, 5, 8, 1, 1, 0, 1, "pac=???"),
  Text(800, 200, 5, 8, 1, 1, 0, 7, "ref=??")
};

Pin ccc_ff2223_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100, 1800,    0, 1800, 1),
  Pin(100, 1600,    0, 1600, 1),
  Pin(100, 1400,    0, 1400, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100, 1000,    0, 1000, 1),
  Pin(100,  800,    0,  800, 1),
  Pin(100,  600,    0,  600, 1),
  Pin(100,  400,    0,  400, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff2223("ccc_ff2223", "ff2223", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    12, ccc_ff2223_pins);

/********************** FF1002  *********************/

Pin ccc_ff1002_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100,  400,    0,  400, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff1002("ccc_ff1002", "ff1002", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    6, ccc_ff1002_pins);

/********************** FF1003  *********************/

Pin ccc_ff1003_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100,  600,    0,  600, 1),
  Pin(100,  400,    0,  400, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff1003("ccc_ff1003", "ff1003", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    7, ccc_ff1003_pins);

/********************** FF1011  *********************/

Pin ccc_ff1011_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100,  800,    0,  800, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff1011("ccc_ff1011", "ff1011", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    6, ccc_ff1011_pins);

/********************** FF1111  *********************/

Pin ccc_ff1111_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100, 1600,    0, 1600, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100,  800,    0,  800, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff1111("ccc_ff1111", "ff1111", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    7, ccc_ff1111_pins);

/********************** FF1112  *********************/

Pin ccc_ff1112_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100, 1600,    0, 1600, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100,  800,    0,  800, 1),
  Pin(100,  400,    0,  400, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff1112("ccc_ff1112", "ff1112", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    8, ccc_ff1112_pins);

/********************** FF1113  *********************/

Pin ccc_ff1113_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100, 1600,    0, 1600, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100,  800,    0,  800, 1),
  Pin(100,  600,    0,  600, 1),
  Pin(100,  400,    0,  400, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff1113("ccc_ff1113", "ff1113", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    9, ccc_ff1113_pins);

/********************** FF1201  *********************/

Pin ccc_ff1201_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100, 1600,    0, 1600, 1),
  Pin(100, 1400,    0, 1400, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff1201("ccc_ff1201", "ff1201", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    7, ccc_ff1201_pins);

/********************** FF1212  *********************/

Pin ccc_ff1212_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100, 1600,    0, 1600, 1),
  Pin(100, 1400,    0, 1400, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100,  800,    0,  800, 1),
  Pin(100,  400,    0,  400, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff1212("ccc_ff1212", "ff1212", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    9, ccc_ff1212_pins);

/********************** FF1222  *********************/

Pin ccc_ff1222_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100, 1600,    0, 1600, 1),
  Pin(100, 1400,    0, 1400, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100, 1000,    0, 1000, 1),
  Pin(100,  800,    0,  800, 1),
  Pin(100,  400,    0,  400, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff1222("ccc_ff1222", "ff1222", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    10, ccc_ff1222_pins);

/********************** FF1223  *********************/

Pin ccc_ff1223_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100, 1600,    0, 1600, 1),
  Pin(100, 1400,    0, 1400, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100, 1000,    0, 1000, 1),
  Pin(100,  800,    0,  800, 1),
  Pin(100,  600,    0,  600, 1),
  Pin(100,  400,    0,  400, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff1223("ccc_ff1223", "ff1223", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    11, ccc_ff1223_pins);

/********************** FF0102nc  *********************/

Pin ccc_ff0102nc_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 1400,    0, 1400, 1),
  Pin(100,  400,    0,  400, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff0102nc("ccc_ff0102nc", "ff0102nc", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    5, ccc_ff0102nc_pins);

/********************** FF1002nc  *********************/

Pin ccc_ff1002nc_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 2000,    0, 2000, 1),
  Pin(100,  400,    0,  400, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff1002nc("ccc_ff1002nc", "ff1002nc", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    5, ccc_ff1002nc_pins);

/********************** FF1111nc  *********************/

Pin ccc_ff1011nc_pins[] =
  {
    Pin(900, 2000, 1000, 2000, 1),
    Pin(900,  400, 1000,  400, 1),
    Pin(100, 2000,    0, 2000, 1),
    Pin(100,  800,    0,  800, 1),
    Pin(100,  200,    0,  200, 1)
  };

Component ccc_ff1011nc("ccc_ff1011nc", "ff1011nc", Component::GATE, 500, 1200,
                       4, ccc_ff_texts,
                       5, ccc_ff1011nc_pins);

/********************** FF1111nc  *********************/

Pin ccc_ff1111nc_pins[] =
  {
    Pin(900, 2000, 1000, 2000, 1),
    Pin(900,  400, 1000,  400, 1),
    Pin(100, 2000,    0, 2000, 1),
    Pin(100, 1600,    0, 1600, 1),
    Pin(100,  800,    0,  800, 1),
    Pin(100,  200,    0,  200, 1)
  };

Component ccc_ff1111nc("ccc_ff1111nc", "ff1111nc", Component::GATE, 500, 1200,
                       4, ccc_ff_texts,
                       6, ccc_ff1111nc_pins);

/********************** FF1212nc  *********************/

// Pin ccc_ff1212nc_pins[] =
// {
//   Pin(900, 2000, 1000, 2000, 1),
//   Pin(900,  400, 1000,  400, 1),
//   Pin(100, 2000,    0, 2000, 1),
//   Pin(100, 1600,    0, 1600, 1),
//   Pin(100, 1400,    0, 1400, 1),
//   Pin(100,  800,    0,  800, 1),
//   Pin(100,  400,    0,  400, 1),
//   Pin(100,  200,    0,  200, 1)
// };

// Component ccc_ff1212nc("ccc_ff1212nc", "ff1212nc", Component::GATE, 500, 1200,
//                  4, ccc_ff_texts,
//                  8, ccc_ff1212nc_pins);

/********************** FF0221  *********************/

Pin ccc_ff0221_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 1600,    0, 1600, 1),
  Pin(100, 1400,    0, 1400, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100, 1000,    0, 1000, 1),
  Pin(100,  800,    0,  800, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff0221("ccc_ff0221", "ff0221", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    8, ccc_ff0221_pins);

/********************** FF0212  *********************/

Pin ccc_ff0212_pins[] =
{
  Pin(900, 2000, 1000, 2000, 1),
  Pin(900,  400, 1000,  400, 1),
  Pin(100, 1600,    0, 1600, 1),
  Pin(100, 1400,    0, 1400, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100,  800,    0,  800, 1),
  Pin(100,  400,    0,  400, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_ff0212("ccc_ff0212", "ff0212", Component::GATE, 500, 1200,
                    4, ccc_ff_texts,
                    8, ccc_ff0212_pins);

/********************** TITLE *********************/
Text ccc_title_texts[] =
{
  Text(23700,  400, 5, 16, 1, 1, 0, 5, "drawing=7002????"),
  Text(18100,  900, 5,  8, 1, 1, 0, 1, "drawn=D. HAMEL"),
  Text(19200,  250, 5,  8, 1, 1, 0, 1, "project=55202"),
  Text(25350,  200, 5, 16, 1, 1, 0, 3, "revision=?"),
  Text(22000, 1400, 5, 10, 1, 1, 0, 0, "title=????"),
  Text(20600,  700, 5, 10, 1, 1, 0, 1, "eng-date=??" "-??" "-??"),
  Text(20600,  500, 5, 10, 1, 1, 0, 1, "app-date=??" "-??" "-??"),
  Text(25100,21700, 5, 10, 1, 1, 0, 5, "lbd-no=?.???")
};
Component ccc_title("ccc_title", "title", Component::TITLE, 0, 0,
                    8, ccc_title_texts,
                    0, 0);

/********************** CC344  *********************/

Text ccc_cc344_texts[] =
{
  Text(500,1200, 5, 8, 1, 1, 0, 4, "refdes=??????"),
  Text(500,2000, 5, 8, 1, 1, 0, 4, "location=???"),
};

Pin ccc_cc344a_pins[] =
{
  Pin(900, 1700, 1000, 1700, 1),  
};

Pin ccc_cc344b_pins[] =
{
  Pin(900, 1700, 1000, 1700, 1),  
  Pin(100, 1700,    0, 1700, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100,  200,    0,  200, 1)
};

Pin ccc_cc344c_pins[] =
{
  Pin(900, 1700, 1000, 1700, 1),  
  Pin(100, 1700,    0, 1700, 1),
  Pin(100, 1200,    0, 1200, 1),
  Pin(100,  700,    0,  700, 1),
  Pin(100,  200,    0,  200, 1)
};

Component ccc_cc344a("ccc_cc344a", "cc344a", Component::GATE, 500, 1700,
                    2, ccc_cc344_texts,
                    1, ccc_cc344a_pins);

Component ccc_cc344b("ccc_cc344b", "cc344b", Component::GATE, 500, 1700,
                    2, ccc_cc344_texts,
                    4, ccc_cc344b_pins);

Component ccc_cc344c("ccc_cc344c", "cc344c", Component::GATE, 500, 1700,
                    2, ccc_cc344_texts,
                    5, ccc_cc344c_pins);

Component ccc_cc344d("ccc_cc344d", "cc344d", Component::GATE, 500, 1700,
                    2, ccc_cc344_texts,
                    5, ccc_cc344c_pins);


Component *Component::components[] = 
{
  &ccc_nand1,
  &ccc_nand2,
  &ccc_nand2a,
  &ccc_nand2b,
  &ccc_nand3,
  &ccc_nand4,
  &ccc_nand5,
  &ccc_nand6,
  &ccc_nand7,
  &ccc_nand8,
  &ccc_nand9,
  &ccc_nand1n,
  &ccc_nand2n,
  &ccc_nand3n,
  &ccc_nand4n,
  &ccc_nand5n,
  &ccc_nand6n,
  &ccc_nand7n,
  &ccc_and1,
  &ccc_and2,
  &ccc_and2a,
  &ccc_and3,
  &ccc_and4,
  &ccc_and5,
  &ccc_and9,
  &ccc_ff2223,
  &ccc_ff1002,
  &ccc_ff1003,
  &ccc_ff1011,
  &ccc_ff1111,
  &ccc_ff1112,
  &ccc_ff1113,
  &ccc_ff1201,
  &ccc_ff1212,
  &ccc_ff1222,
  &ccc_ff1223,
  &ccc_ff0102nc,
  &ccc_ff1002nc,
  &ccc_ff1011nc,
  &ccc_ff1111nc,
  //&ccc_ff1212nc,
  &ccc_ff0221,
  &ccc_ff0212,
  &ccc_title,
  
  &ccc_cc344a,
  &ccc_cc344b,
  &ccc_cc344c,
  &ccc_cc344d,
  
  0
};

Component *Component::lookup(const char *nickname)
{
  int i = 0;

  while (components[i])
    {
      if (strcmp(components[i]->nickname, nickname)==0)
        return components[i];
      i++;
    }
  return 0;
}
