/*
 * makesch - make a skeleton schematic from text file
 *
 * Copyright (C) 2006, 2007  Adrian Wise
 *
 * This file is part of makesch.
 *
 * makesch is free software: you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See http://www.series16.adrianwise.co.uk
 *   mailto:apw@adrianwise.co.uk
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "components.hh"

char *readstr(FILE *fp)
{
  char buf[1024];
  int i=0;
  int c;
  
  /*
   * Discard leading white space
   */
  
  do
    {
      c = getc(fp);
    } while (isspace(c) && (c!='\n') && (c!=EOF));
  
  if (c==EOF)
    return NULL;
  else if (c!='\n')
    {
      if (c == '"')
        {
          c = getc(fp);
          while ((c != '"') && (c != '\n') && (c != EOF))
            {
              buf[i++] = c;
              c = getc(fp);
            }
        }
      else
        {
          do
            {
              buf[i++] = c;
              c = getc(fp);
            } while((!isspace(c)) && (c!=EOF));
          ungetc(c, fp);
        }
      
      buf[i] = '\0';
    }
  
  char *r = new char[i+1];
  int j;
  for (j=0; j<i; j++)
    r[j] = buf[j];
  r[j] = '\0';
  
  return r;
}

char **readline(FILE *fp)
{
  char *args[1024];
  int a=0;
  char *str=readstr(fp);
  
  if (!str)
    return NULL; // EOF
  
  while ((str) && (strlen(str)>0))
    {
      //printf("str = \"%s\"\n", str);
      args[a++] = str;
      str=readstr(fp);
    } 
  
  if (str)
    delete [] str; // The newline string
  
  char **r = new char *[a+1];
  int i;
  for (i=0; i<a; i++)
    r[i] = args[i];
  r[i] = 0;
  
  return r;
}

void deleteline(char **line)
{
  int i=0;
  
  while (line[i])
    {
      delete [] line[i];
      i++;
    }
  
  delete [] line;
}

void printline(FILE *fp, char **line)
{
  int i=0;
  while (line[i])
    {
      fprintf(fp, "%s ", line[i]);
      i++;
    }
  fprintf(fp, "\n");
}

char ***readpara(FILE *fp)
{
  char **lines[1024];
  int l=0;
  char **line;
  
  /*
   * Discard leading empty lines
   */
  
  do
    {
      line = readline(fp);
                } while (line && (!line[0]));
  
  if (!line)
    return NULL; // EOF
  
  while (line && line[0])
    {
      //printline(line);
      lines[l++] = line;
      line=readline(fp);
    } 
  
  if (line)
    deleteline(line); // The closing empty line
  
  char ***r = new char **[l+1];
  int i;
  for (i=0; i<l; i++)
    r[i] = lines[i];
  r[i] = 0;
  
  return r;
}

void deletepara(char ***para)
{
  int i=0;
  
  while (para[i])
    {
      deleteline(para[i]);
      i++;
    }
  
  delete [] para;
}

void printpara(FILE *fp, char ***para)
{
  int i=0;
  while (para[i])
    {
      printline(fp, para[i]);
      i++;
    }
  fprintf(fp, "\n");
}


void expect_words(int words, char **line)
{
  int i = 0;
  
  while (line[i])
    i++;
  
  if (i != words)
    {
      printline(stderr, line);
      fprintf(stderr, "Expected %d words, found %d\n", words, i);
      exit(1);
    }
}

void expect_lines(int lines, char ***para)
{
  int i = 0;
  
  while (para[i])
    i++;
  
  if (i != lines)
    {
      fprintf(stderr, "Expected %d lines, found %d\n", lines, i);
      printpara(stderr, para);
      exit(1);
    }
  
}

void parse_location(char *str, int &x, int &y)
{
  int c = toupper(str[0]);
  int xx, yy;
  
  if ((c < 'A') || (c > 'P') || (c == 'I') || (c == 'O'))
    {
      fprintf(stderr, "<%c> is an illegal horiz. location\n", c);
      exit(1);
    }
  else
    xx = (c < 'I') ? (c - 'A') :
    (c < 'O') ? (c - 'B') : (c - 'C');
  
  char *ptr = &str[1];
  yy = strtol(&str[1], &ptr, 10);
  if (ptr == &str[1])
    {
      fprintf(stderr, "No integer in <%s>\n", &str[1]);
      exit(1);
    }
  if ((yy<1) || (yy>12))
    {
      fprintf(stderr, "<%d> is an illegal vertical location\n", yy);
      exit(1);
    }
  yy = (12-yy);
  
  x = 2500 + (xx * 1600);
  y = 3100 + (yy * 1600);
  
  //printf("Coords (%d, %d)\n", x, y);
}

void instance(char ***para)
{
  Component *component;
  int pins;
  enum Component::TYPE type; 
  int x, y;
  int i;
  
  component = Component::lookup(para[0][0]);
  if (!component)
    {
      fprintf(stderr, "Unknown component <%s>\n", para[0][0]);
      exit(1);
    }

  pins = component->get_num_pins();
  type = component->get_type();

  if (type == Component::GATE)
    {
      expect_lines(1+pins, para);
      expect_words(2+component->get_num_texts(), para[0]);
      parse_location(para[0][1], x, y);
      
      component->instance(x, y, &para[0][2]);
      
      //printf("pins = %d\n", pins);
      
      for (i=1; i<=pins; i++)
        {
          expect_words(3, para[i]);
          char *signal;
          char *page;
          char *post;
          
          signal = (strcmp(para[i][0], "*")==0) ? 0 : para[i][0];
          page = (strcmp(para[i][1], "*")==0) ? 0 : para[i][1];
          post = (strcmp(para[i][2], "*")==0) ? 0 : para[i][2];
          
          component->annotate(x, y, i-1, signal, page, post);
        }
    }
  else if (type == Component::TITLE)
    {
      expect_lines(1, para);
      expect_words(9, para[0]);
      
      component->instance(0, 0, &para[0][1]);
    }
  else
    abort();
}

int main(int argc, char **argv)
{
  if (argc != 2)
    {
      fprintf(stderr, "usage: %s <filename>\n", argv[0]);
      exit(1);
    }
  
  FILE *fp = fopen(argv[1], "r");
  if (!fp)
    {
      fprintf(stderr, "Could not open <%s> for reading\n", argv[1]);
      exit(1);
    }
  
  printf("v 20030525\n");
  
  char ***para = readpara(fp);
  
  while (para)
    {
      //printf("para = \n");
      //printpara(para);
      
      instance(para);
      
      deletepara(para);
      para = readpara(fp);
    }

  fclose(fp);
  exit(0);
}
