/***********************************************************************
 *
 * h316 - Memory
 * 
 * Copyright (C) 2007, 2011, 2017 Adrian Wise
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See http://www.series16.adrianwise.co.uk
 *   mailto:apw@adrianwise.co.uk
 * 
 ***********************************************************************
 * 
 * This models memory connected to a gate-level h316 CPU.
 * 
 * A single 32K word memory is implemented, but two separate interfaces
 * are provided to it - one for the lower 16K and the other to the
 * upper 16K. This provides for both the standard 16K machine, and the
 * memory expansion option. Only one or the other interface can be
 * active at any one time.
 * 
 * On the real machine, each 4K would have been provided by a
 * separate memory module - each one addressed by its own BANK signal.
 * In this model the BANK signals are decoded back into address bits,
 * and the active interface (standard or extended memory) provides the
 * most significant address bit.
 * 
 * It is interesting to note that the lower address bits (addressing
 * within a 4K module) are of opposite polarity on the standard memory
 * interface and the extended memory interface.
 * 
 **********************************************************************/

module h316_memory (/*AUTOARG*/
   // Outputs
   MMxxF_n, MMDxx_n,
   // Inputs
   MBSYX_p, XTIMG_p, YTIMG_p, WRITE_n, STROB_n, MADCL_n_A, MADCL_n_C,
   BANKA_n, BANKB_n, BANKC_n, BANKD_n, MADxx_n, MxxFF_p, MBUSY_p,
   MXTMG_p, MYTMG_p, xWRITE_n, MSTRB_n, MARRC_n_A, MARRC_n_B,
   xBANKA_n, xBANKB_n, xBANKC_n, xBANKD_n, xMADxx_n, MDOxx_p
   );

   // 1st 16K
   input MBSYX_p; // Memory cycle
   input XTIMG_p;
   input YTIMG_p;
   input WRITE_n;
   input STROB_n;
   input MADCL_n_A;
   input MADCL_n_C;
   input BANKA_n;
   input BANKB_n;
   input BANKC_n;
   input BANKD_n;
   input [5:16] MADxx_n;
   input [1:16] MxxFF_p;
   output wire [1:16] MMxxF_n;

   // Memory Expansion
   input MBUSY_p;
   input MXTMG_p;
   input MYTMG_p;
   input xWRITE_n;
   input MSTRB_n;
   input MARRC_n_A;
   input MARRC_n_B;
   input xBANKA_n;
   input xBANKB_n;
   input xBANKC_n;
   input xBANKD_n;
   input [5:16] xMADxx_n;
   input [1:16] MDOxx_p;
   output wire [1:16] MMDxx_n;

   reg [1:16]    memory [0:32767];
   reg [2:16]    addr;
   integer       a;
   reg [1:16]    rdata;
 
   wire          mbusy;
   wire          xtimg;
   wire          wrt_n;
   wire [1:16]   mdata;
   
   reg [256*8-1:1] memfilename;

   initial begin
      for (a = 0; a<32768; a=a+1) begin
         memory[a] <= 16'd0;
      end

      #1;
      
      if ($value$plusargs("MEMFILE=%s", memfilename)) begin
         $display("About to read %0s into gate sim", memfilename);
         $readmemh(memfilename, memory);
      end else begin
         $display("MEMFILE not set");
         $finish_and_return(1);
      end
            
   end // initial begin
   
   //
   // Address latch
   //
   always @(/*AUTOSENSE*/BANKA_n or BANKB_n or BANKC_n or BANKD_n
            or MADxx_n or MBSYX_p or MBUSY_p or mbusy or xBANKA_n
            or xBANKB_n or xBANKC_n or xBANKD_n or xMADxx_n) begin
      if (~mbusy)
        addr = 15'o77777;
      else begin
         if (MBSYX_p) begin
            // MADxx_n is actually +ve logic address
            addr[5:16] = addr[5:16] & MADxx_n[5:16];
            case ( {BANKA_n, BANKB_n, BANKC_n, BANKD_n} )
              4'b1111: ; // Do nothing
              4'b0111: addr[3:4] = 2'b00;
              4'b1011: addr[3:4] = 2'b01;
              4'b1101: addr[3:4] = 2'b10;
              4'b1110: addr[3:4] = 2'b11;
              default: addr[3:4] = 2'bxx;
            endcase // case ( {BANKA_n, BANKB_n, BANKC_n, BANKD_n} )
            addr[2] = 1'b0;
         end else if (MBUSY_p) begin // if (MBSYX_p)
            // But xMADxx_n is inverted (as its name suggests)
            if (!MXTMG_p) begin
               addr[5:16] = 12'd0;
            end
            addr[5:16] = addr[5:16] | ~xMADxx_n[5:16];
            case ( {xBANKA_n, xBANKB_n, xBANKC_n, xBANKD_n} )
              4'b1111: ; // Do nothing
              4'b0111: addr[3:4] = 2'b00;
              4'b1011: addr[3:4] = 2'b01;
              4'b1101: addr[3:4] = 2'b10;
              4'b1110: addr[3:4] = 2'b11;
              default: addr[3:4] = 2'bxx;
            endcase // case ( {xBANKA_n, xBANKB_n, xBANKC_n, xBANKD_n} )
            addr[2] = 1'b1;
         end // if (MBUSY_p)
      end // else: !if(~MBSYX_p || ~MBUSY_p)
   end // always @ (...

   assign mbusy =             MBSYX_p |  MBUSY_p;
   assign xtimg = (MBSYX_p) ? XTIMG_p :  MXTMG_p;
   assign wrt_n = (MBSYX_p) ? WRITE_n : xWRITE_n;
   assign mdata = (MBSYX_p) ? MxxFF_p :  MDOxx_p;
   
   //
   // Read and write memory
   //
   always @(posedge xtimg) begin
      if (mbusy) begin
         a = addr[2:16];  
         if (wrt_n) begin
            // Do a read
            if (^addr === 1'bx) begin
               rdata <= 16'bx;
            end else begin
               rdata <= memory[a];
               memory[a] <= 16'bxxxxxxxxxxxxxxxx;
            end
         end else begin // if (WRITE_n)
            // Do a write
            memory[a] <= mdata;
            //$display("write: memory[", a,"] = ", mdata);
         end // else: !if(WRITE_n)
      end // if (MBSYX_p)
   end // always @ (posedge XTIMG_p)
   
   // Output buffers
   nand (strong0, pull1) #16 obuf01 (MMxxF_n[ 1], rdata[ 1], ~STROB_n );
   nand (strong0, pull1) #16 obuf02 (MMxxF_n[ 2], rdata[ 2], ~STROB_n );
   nand (strong0, pull1) #16 obuf03 (MMxxF_n[ 3], rdata[ 3], ~STROB_n );
   nand (strong0, pull1) #16 obuf04 (MMxxF_n[ 4], rdata[ 4], ~STROB_n );
   nand (strong0, pull1) #16 obuf05 (MMxxF_n[ 5], rdata[ 5], ~STROB_n );
   nand (strong0, pull1) #16 obuf06 (MMxxF_n[ 6], rdata[ 6], ~STROB_n );
   nand (strong0, pull1) #16 obuf07 (MMxxF_n[ 7], rdata[ 7], ~STROB_n );
   nand (strong0, pull1) #16 obuf08 (MMxxF_n[ 8], rdata[ 8], ~STROB_n );
   nand (strong0, pull1) #16 obuf09 (MMxxF_n[ 9], rdata[ 9], ~STROB_n );
   nand (strong0, pull1) #16 obuf10 (MMxxF_n[10], rdata[10], ~STROB_n );
   nand (strong0, pull1) #16 obuf11 (MMxxF_n[11], rdata[11], ~STROB_n );
   nand (strong0, pull1) #16 obuf12 (MMxxF_n[12], rdata[12], ~STROB_n );
   nand (strong0, pull1) #16 obuf13 (MMxxF_n[13], rdata[13], ~STROB_n );
   nand (strong0, pull1) #16 obuf14 (MMxxF_n[14], rdata[14], ~STROB_n );
   nand (strong0, pull1) #16 obuf15 (MMxxF_n[15], rdata[15], ~STROB_n );
   nand (strong0, pull1) #16 obuf16 (MMxxF_n[16], rdata[16], ~STROB_n );

   nand (strong0, pull1) #16 xbuf01 (MMDxx_n[ 1], rdata[ 1], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf02 (MMDxx_n[ 2], rdata[ 2], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf03 (MMDxx_n[ 3], rdata[ 3], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf04 (MMDxx_n[ 4], rdata[ 4], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf05 (MMDxx_n[ 5], rdata[ 5], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf06 (MMDxx_n[ 6], rdata[ 6], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf07 (MMDxx_n[ 7], rdata[ 7], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf08 (MMDxx_n[ 8], rdata[ 8], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf09 (MMDxx_n[ 9], rdata[ 9], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf10 (MMDxx_n[10], rdata[10], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf11 (MMDxx_n[11], rdata[11], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf12 (MMDxx_n[12], rdata[12], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf13 (MMDxx_n[13], rdata[13], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf14 (MMDxx_n[14], rdata[14], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf15 (MMDxx_n[15], rdata[15], ~MSTRB_n );
   nand (strong0, pull1) #16 xbuf16 (MMDxx_n[16], rdata[16], ~MSTRB_n );
   
endmodule // h316_memory
