/*
 * H316 DMC verification component.
 * 
 * Copyright (C) 2014, 2015, 2017 Adrian Wise
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See http://www.series16.adrianwise.co.uk
 *   mailto:apw@adrianwise.co.uk
 *
 ***********************************************************************
 * 
 * This models a "peripheral" that may be used to exercise the
 * DMC controller of an H16 CPU. It is purely a verfification
 * component - it has no application in a real system.
 * 
 * It consists of a central control block together with upto
 * 16 (NDMC) channel exerciser blocks, h16_vdmc_chn, that may
 * be used to generate and check DMC traffic. The central
 * controller provides a register to select the currently
 * active controller to which a series of per-channel INA,
 * OTA, SKS, and OCP operations are directed.
 * 
 * Each h16_vdmc_chn channel exerciser can operate in either
 * input mode or output mode.
 * 
 * In input mode it generates data words using a defined
 * pseudo random number generator which are written into memory
 * under DMC control. A checksum is added as the last word of
 * the transer which can then be re-computed by the CPU to
 * determine if the transfer was successful.
 * 
 * In output mode it accepts data words from memory under
 * DMC control. It both checks these words against the PRNG and
 * calculates the checksum over these words allowing two ways
 * to determine if the transfer was successful. (If the CPU
 * generates the data using the PRNG, then every word can be
 * checked, if it uses other data then the checksum can still
 * be checked.)
 * 
 * In either input or output mode a second PRNG may be used to
 * determine how long, in clock cycles, are left between DMC
 * requests. Alternatively a configurable but, during a given
 * transfer, constant time delay may be used.
 *
 * This peripheral uses two I/O device addresses. The first
 * (xx below, with default address '37) is used by the central
 * control block. The second (yy below, with default address '36)
 * addresses the currently selected channel.
 * 
 * I/O operations to the central controller:
 * 
 * OTA '07xx Output A to channel select register (4 LSBs)
 * INA '17xx Clear A and input channel select register
 * INA '16xx Clear A and input highest priority interrupting chn.
 * INA '15xx Clear A and input highest priority chn signalling
 *           an unexpected transfer error.
 * 
 * SKS '01xx - Skip if no channels are busy
 * SKS '03xx - Skip if no channels are signalling an unexpected
 *             transfer error.
 * SKS '04xx - Skip if not interrupting
 * OCP '03xx - Clear all channels' unexpected transfer error.
 * 
 * Values in the channel select register (0 to 15) are one
 * smaller than DMC channel number (1 to 16).
 * 
 * I/O operations to the currently selected channel exerciser:
 * 
 * OTA '00yy Output A to transfer size register
 * INA '10yy Clear A and input from transfer size register
 * 
 * OTA '02yy Output A to data word PNRG seed register
 * INA '12yy Clear A and input from data word PNRG seed register
 * 
 * OTA '05yy Output A to timing PNRG seed register
 * INA '15yy Clear A and input from timing PNRG seed register
 * 
 * OTA '06yy Output A to timing control register
 * INA '16yy Clear A and input from timing control register
 * 
 * Format of the Timing control register:
 * 
 * M MMM NNN NNN NNN NNN
 * 
 * T = NNNNNNNNNNNN + (PRNG & ((1 << MMMM)-1))
 * 
 * If (MMMM == 0) the delay is the constant given in the lower
 * 12 bits, else a variable random component to the delay.
 * 
 * INA '17yy Clear A and input error bits
 * 
 * Lower 5 bits contain error indications:
 * 4: Unexpected Transfer
 * 3: Missing End-Of-Range transfer
 * 2: Unexpected End-Of-Range transfer
 * 1: A data word mismatched the PRNG
 * 0: The checksum is non-zero
 * 
 * OCP '03yy - Clear transfer errors.
 * OCP '04yy - Clear channel's interrupt flop
 * OCP '05yy - Start DMC input transfer
 * OCP '06yy - Start DMC output transfer
 *
 * SKS '01yy - Skip if not busy
 * SKS '03yy - Skip if no error condition
 * SKS '04yy - Skip if not interrupting
 * 
 * vdmc will generate an interrupt at end of DMC transfer.
 */

module h316_vdmc_chn
  #(parameter CHAN=-1)
   (// Asynchronous CPU I/O-bus interface
    input  wire        addressed,       // This channel addressed
    input  wire  [1:4] function_code,   // Function code (address bus)
    output reg         chan_cpu_drlin,  // Ready line
    output reg  [15:0] chan_cpu_data,   // Read data from channel
    input  wire        DALxx_n,         // DMC acknowledge 
    // Synchonised CPU I/O-bus
    input  wire        clk,             // Clock
    input  wire        reset_N,         // Reset
    input  wire        reg_addressed,   // This channel addressed 
    input  wire  [3:0] reg_fcode,       // Device function code 
    input  wire [15:0] reg_OTBxx,       // Output bus
    input  wire        reg_RRLIN,       // INA/OTA/SKS cycle
    input  wire        reg_OCPLS,       // OCP cycle
    input  wire        reg_DALxx,       // DMC cycle
    input  wire        reg_ERLXX,       // DMC end-of-range
    output reg         DILxx_p,         // DMC request
    // Interface to central controller
    output reg         unexpected_trfr, // Unexpected transfer
    input wire         clear_error,     // Clear unexpected transfer error
    output wire        busy,            // Channel busy
    output reg         chan_interrupt   // Channel interrupt
    );   
   
   localparam OTA_TRFR_SIZE = 4'o00;
   localparam SKS_NOT_BUSY  = 4'o01;
   localparam OTA_DATA_PRNG = 4'o02;
   localparam SKS_NO_ERROR  = 4'o03;
   localparam SKS_NOT_INTR  = 4'o04;
   localparam OTA_TIME_PRNG = 4'o05;   
   localparam OTA_TIME_CTRL = 4'o06;
   //                            07
   localparam INA_TRFR_SIZE = 4'o10;
   //                            11
   localparam INA_DATA_PRNG = 4'o12;
   //                            13
   //                            14
   localparam INA_TIME_PRNG = 4'o15;
   localparam INA_TIME_CTRL = 4'o16;
   localparam INA_CHAN_ERRS = 4'o17;

   localparam OCP_CLEAR_ERR = 4'o03;
   localparam OCP_CLEAR_INT = 4'o04;
   localparam OCP_STRT_INPT = 4'o05;
   localparam OCP_STRT_OUTP = 4'o06;

   localparam PRNG_TAPS = 16'hA55A;
   
   reg [15:0]          trfr_size, next_trfr_size;
   reg [15:0]          data_prng, next_data_prng;
   reg [15:0]          time_prng, next_time_prng;
   reg [15:0]          time_ctrl, next_time_ctrl;
   reg [15:0]          time_cntr, next_time_cntr;
   reg                 timing, next_timing;
   reg                 reload_time_cntr;
   reg [15:0]          trfr_cntr, next_trfr_cntr;
   reg                 increment_pending_transfers;
   reg                 decrement_pending_transfers;
   reg [15:0]          pending_transfers, next_pending_transfers;
   reg                 last_pending, next_last_pending;
   reg                 last_transfer;
   reg                 input_mode, next_input_mode;
   reg [15:0]          checksum, next_checksum;
   reg                 next_chan_interrupt;
   reg                 data_mismatch, next_data_mismatch;  
   reg                 unexpected_eor, next_unexpected_eor;
   reg                 missing_eor, next_missing_eor;
   reg                 next_unexpected_trfr;
   reg                 start;
   reg [15:0]          data;
   
   wire [4:0]          error_bits;

   always @(/*AUTOSENSE*/checksum or data_prng or input_mode
            or last_transfer) begin
      if (input_mode) begin
         if (last_transfer) begin
            data = 16'd0 - {checksum[0], checksum[15:1]};
         end else begin
            data = data_prng;
         end
      end else begin
         data = 16'd0;
      end
   end
   
   // Drive device-ready line and input bus
   always @(/*AUTOSENSE*/DALxx_n or addressed or busy
            or chan_interrupt or data or data_prng or error_bits
            or function_code or time_ctrl or time_prng or trfr_size) begin
      
      chan_cpu_drlin =  1'b0;
      chan_cpu_data  = 16'd0;

      if (!DALxx_n) begin
         chan_cpu_data = data;
      end else if (addressed) begin
         case (function_code)
           OTA_TRFR_SIZE: chan_cpu_drlin = 1'b1;
           SKS_NOT_BUSY:  chan_cpu_drlin = ~busy;
           OTA_DATA_PRNG: chan_cpu_drlin = 1'b1;
           SKS_NO_ERROR:  chan_cpu_drlin = ~|error_bits;
           SKS_NOT_INTR:  chan_cpu_drlin = ~chan_interrupt;
           OTA_TIME_PRNG: chan_cpu_drlin = 1'b1;
           OTA_TIME_CTRL: chan_cpu_drlin = 1'b1;
           
           INA_TRFR_SIZE: begin chan_cpu_drlin = 1'b1; chan_cpu_data = trfr_size; end
           INA_DATA_PRNG: begin chan_cpu_drlin = 1'b1; chan_cpu_data = data_prng; end
           INA_TIME_PRNG: begin chan_cpu_drlin = 1'b1; chan_cpu_data = time_prng; end
           INA_TIME_CTRL: begin chan_cpu_drlin = 1'b1; chan_cpu_data = time_ctrl; end
           INA_CHAN_ERRS: begin chan_cpu_drlin = 1'b1; chan_cpu_data = {11'd0, error_bits}; end

           default: begin /* Do Nothing */ end
         endcase // case (function_code)
      end // if (addressed)
   end // always @ (...

   always @(/*AUTOSENSE*/DILxx_p or chan_interrupt or checksum
            or clear_error or data or data_mismatch or data_prng
            or input_mode or last_pending or missing_eor
            or pending_transfers or reg_DALxx or reg_ERLXX
            or reg_OCPLS or reg_OTBxx or reg_RRLIN or reg_addressed
            or reg_fcode or time_cntr or time_ctrl or time_prng
            or timing or trfr_cntr or trfr_size or unexpected_eor
            or unexpected_trfr) begin

      next_trfr_size = trfr_size;
      next_data_prng = data_prng;
      next_time_prng = time_prng;
      next_time_ctrl = time_ctrl;
      next_time_cntr = time_cntr;
      next_timing    = timing;
      next_trfr_cntr = trfr_cntr;
      next_pending_transfers = pending_transfers;
      next_last_pending = last_pending;
      next_input_mode = input_mode;
      next_chan_interrupt = chan_interrupt;
      next_checksum = checksum;
      next_data_mismatch = data_mismatch;
      next_unexpected_eor = unexpected_eor;
      next_missing_eor = missing_eor;
      next_unexpected_trfr = unexpected_trfr;
      start = 1'b0;
      reload_time_cntr = 1'b0;
      increment_pending_transfers = 1'b0;
      decrement_pending_transfers = 1'b0;
      
      if (reg_addressed) begin
         if (reg_OCPLS) begin // OCP
            
            case (reg_fcode)
              OCP_CLEAR_ERR: begin
                 // Clear channel errors
                 next_data_mismatch  = 1'b0;
                 next_unexpected_eor = 1'b0;
                 next_missing_eor    = 1'b0;
              end
              OCP_CLEAR_INT: next_chan_interrupt = 1'b0;
              OCP_STRT_OUTP: begin next_input_mode = 1'b0; start = 1'b1; end
              OCP_STRT_INPT: begin next_input_mode = 1'b1; start = 1'b1; end
              default:       begin /* Do nothing */ end
            endcase // case (reg_fcode)

         end else if (reg_RRLIN) begin // if (reg_OCPLS)
            // INA/OTA/SKS
            
            case (reg_fcode)
              OTA_TRFR_SIZE: begin next_trfr_size = reg_OTBxx; end
              OTA_DATA_PRNG: begin next_data_prng = reg_OTBxx; end
              OTA_TIME_PRNG: begin next_time_prng = reg_OTBxx; end
              OTA_TIME_CTRL: begin next_time_ctrl = reg_OTBxx; end
              default:       begin /* Do nothing */ end
            endcase // case (reg_fcode)
         end // if (reg_RRLIN)
      end // if (reg_addressed)
      
      last_transfer = ((pending_transfers == 16'd1) && (last_pending));
      
      if (reg_DALxx && reg_RRLIN) begin
         // DMC transfer
         
         // Only expect a transfer if requesting
         if (!DILxx_p) begin
            next_unexpected_trfr = 1'b1;
         end
         
         if (reg_ERLXX && (!last_transfer)) begin
            next_unexpected_eor = 1'b1;
         end
         
         if (last_transfer && (!reg_ERLXX)) begin
            next_missing_eor = 1'b1;
         end
         
         decrement_pending_transfers = 1'b1;
         
         if (input_mode) begin

            next_checksum = {checksum[0], checksum[15:1]} + data;

         end else begin
            
            next_checksum = {checksum[0], checksum[15:1]} + reg_OTBxx;
            
            if (!last_transfer) begin
               if (reg_OTBxx != data_prng)
                 next_data_mismatch = 1'b1;
            end
            
         end // else: !if(input_mode)
         
         if (last_transfer) begin
            next_chan_interrupt = 1'b1;                  
         end else begin
            next_data_prng = ((data_prng >> 1) ^ ((data_prng[0]) ? PRNG_TAPS : 0));
         end
         
      end // if (reg_DALxx && reg_RRLIN)
      
      if (start) begin
         reload_time_cntr       = 1'b1;
         next_trfr_cntr         = trfr_size - 16'd1;
         next_pending_transfers = 16'd0;
         next_last_pending      = 1'b0;
         next_checksum          = 16'd0;
         
         next_chan_interrupt    = 1'b0; // Clear interrupt
         
         // Clear channel errors
         next_data_mismatch     = 1'b0;
         next_unexpected_eor    = 1'b0;
         next_missing_eor       = 1'b0;

      end // if (start)

      if ((!reload_time_cntr) && (timing)) begin
         if (time_cntr == 16'd0) begin
            increment_pending_transfers = 1'b1;
            next_timing = 1'b0;
            if (trfr_cntr == 16'd0) begin
               next_last_pending = 1'b1;
            end else begin
               next_trfr_cntr = trfr_cntr - 16'd1;
               reload_time_cntr = 1'b1;
            end
         end else begin
            next_time_cntr = time_cntr - 16'd1;
         end // else: !if(time_cntr == 16'd0)
      end // if ((!reload_time_cntr) && (timing))
      
      if (reload_time_cntr) begin
         
         next_time_cntr = {4'd0, time_ctrl[11:0]};
         
         if (time_ctrl[15:12] != 4'd0) begin
            case (time_ctrl[15:12])
              4'd01: next_time_cntr = next_time_cntr + {15'd0, time_prng[   0]};
              4'd02: next_time_cntr = next_time_cntr + {14'd0, time_prng[ 1:0]};
              4'd03: next_time_cntr = next_time_cntr + {13'd0, time_prng[ 2:0]};
              4'd04: next_time_cntr = next_time_cntr + {12'd0, time_prng[ 3:0]};
              4'd05: next_time_cntr = next_time_cntr + {11'd0, time_prng[ 4:0]};
              4'd06: next_time_cntr = next_time_cntr + {10'd0, time_prng[ 5:0]};
              4'd07: next_time_cntr = next_time_cntr + { 9'd0, time_prng[ 6:0]};
              4'd08: next_time_cntr = next_time_cntr + { 8'd0, time_prng[ 7:0]};
              4'd09: next_time_cntr = next_time_cntr + { 7'd0, time_prng[ 8:0]};
              4'd10: next_time_cntr = next_time_cntr + { 6'd0, time_prng[ 9:0]};
              4'd11: next_time_cntr = next_time_cntr + { 5'd0, time_prng[10:0]};
              4'd12: next_time_cntr = next_time_cntr + { 4'd0, time_prng[11:0]};
              4'd13: next_time_cntr = next_time_cntr + { 3'd0, time_prng[12:0]};
              4'd14: next_time_cntr = next_time_cntr + { 2'd0, time_prng[13:0]};
              4'd15: next_time_cntr = next_time_cntr + { 1'd0, time_prng[14:0]};
              default: begin /* Do nothing */ end
            endcase // case (time_ctrl[15:12])
            
            next_time_prng = ((time_prng >> 1) ^ ((time_prng[0]) ? PRNG_TAPS : 0));
         end // if (time_ctrl[15:12] != 4'd0)
         
         next_timing = 1'b1;
      end // if (reload_time_cntr)
      
      if (increment_pending_transfers && (!decrement_pending_transfers))
        next_pending_transfers = pending_transfers + 16'd1;
      if (decrement_pending_transfers && (!increment_pending_transfers))
        next_pending_transfers = pending_transfers - 16'd1;

      if (clear_error)
        next_unexpected_trfr = 1'b0;
      
   end // always @ (...

   assign busy = timing | (|pending_transfers);
   assign error_bits  = {unexpected_trfr,
                         missing_eor,
                         unexpected_eor,
                         data_mismatch,
                         (checksum != 16'd0)};
   
   always @(posedge clk) begin
      if (!reset_N) begin
         trfr_size         <=  16'd0;
         data_prng         <= ~16'd0;
         time_prng         <= ~16'd0;
         time_ctrl         <=  16'd0;
         time_cntr         <=  16'd0;
         timing            <=   1'b0;
         trfr_cntr         <=  16'd0;
         pending_transfers <=  16'd0;
         last_pending      <=   1'b0;
         DILxx_p           <=   1'b0;
         input_mode        <=   1'b0;
         checksum          <=  16'd0;
         chan_interrupt    <=   1'b0;
         data_mismatch     <=   1'b0;
         unexpected_eor    <=   1'b0;
         missing_eor       <=   1'b0;
         unexpected_trfr   <=   1'b0;
      end else begin // if (!reset_N)
         trfr_size         <= next_trfr_size;
         data_prng         <= next_data_prng;
         time_prng         <= next_time_prng;
         time_ctrl         <= next_time_ctrl;
         time_cntr         <= next_time_cntr;
         timing            <= next_timing;
         trfr_cntr         <= next_trfr_cntr;
         pending_transfers <= next_pending_transfers;
         last_pending      <= next_last_pending;
         DILxx_p           <= (|next_pending_transfers);
         input_mode        <= next_input_mode;
         checksum          <= next_checksum;
         chan_interrupt    <= next_chan_interrupt;
         data_mismatch     <= next_data_mismatch;
         unexpected_eor    <= next_unexpected_eor;
         missing_eor       <= next_missing_eor;
         unexpected_trfr   <= next_unexpected_trfr;

         if (reg_DALxx && reg_RRLIN) begin
            if ( trfr_cntr == 16'd0 ) begin 
               $display("%t: Channel %d: trfr_cntr = %d, last_transfer = %d",
                        $time, CHAN, trfr_cntr, last_transfer);
            end else begin
               $display("%t: Channel %d: trfr_cntr = %d",
                        $time, CHAN, trfr_cntr);
            end
         end
         if (start) begin
            $display("%t: Channel %d start transfer size = %d",
                     $time, CHAN, trfr_size);
         end
         
      end // else: !if(!reset_N)
   end // always @ (posedge clk)

endmodule // h16_vdmc_chn

module h316_vdmc
  #(parameter CTRL_ADDR  = 6'o37, // Central controller device address
    parameter CHAN_ADDR  = 6'o36, // Channel controller device address
    parameter NDMC       = 16,    // Number of channel exercisers
    parameter SMK_BIT    = 3)     // Standard interrupt SMK bit
   (// Modern synchronous clocked logic
    input  wire        clk,       // Clock
    input  wire        reset_N,   // Reset
    // Asynchronous CPU I/O-bus interface
    input  wire        MSTCL_n,   // Master clear
    input  wire [7:16] ADBxx_n,   // Address bus
    input  wire [1:16] OTBxx_p,   // Output bus
    input  wire        OCPLS_n,   // OCP
    input  wire        SMK01_n,   // Set mask strobe
    input  wire        RRLIN_n,   // Reset ready line (INA, OTA)
    output wire [1:16] INBxx_n,   // Input bus
    output wire        DRLIN_n,   // Device ready line
    output wire        PIL00_n,   // Standard interrupt line
    output wire        PILVD_n,   // Priority interrupt line (VDmc)
    // DMC request/acknowledge
    input  wire        ERLXX_n,   // DMC end of range signal
    input  wire [1:16] DALxx_n,   // DMC acknowledge
    output wire [1:16] DILxx_p);  // DMC request

   localparam OTA_CHAN_SLCT = 4'o07;
   
   localparam INA_CHAN_SLCT = 4'o17;
   localparam INA_INTR_CHAN = 4'o16;
   localparam INA_ERR_CHAN  = 4'o15;
   
   localparam SKS_NOT_BUSY  = 4'o01;
   localparam SKS_NTRFR_ERR = 4'o03;
   localparam SKS_NOT_INTR  = 4'o04;
   
   localparam OCP_CLEAR_ERR = 4'o03;

   genvar i;
   
   wire        ctrl_addressed;
   wire        chan_addressed;

   wire  [1:6] device_address;
   wire  [1:4] function_code;
   reg         ctrl_cpu_drlin;
   reg  [15:0] ctrl_cpu_data;

   wire        strobe_N;
   reg         lat_OCPLS_n;
   reg         lat_RRLIN_n;
   reg         lat_ERLXX_n;
   reg         intr_mask;

   reg         reg_OCPLS;
   reg         reg_RRLIN;
   reg         reg_ctrl_addr;
   reg         reg_chan_addr;
   reg   [3:0] reg_fcode;
   reg  [15:0] reg_OTBxx;
   reg         reg_ERLXX;
   reg  [1:16] tmp_DALxx;
   reg         bus_update;
   
   wire        strobe;
   reg  [15:0] reg_DALxx;
   reg  [15:0] addressed;
   reg  [15:0] reg_addressed;

   reg         clk_update, next_clk_update;   
   reg   [3:0] channel,    next_channel;
   reg         clear_error;
   
   // Channel controller outputs
   wire [15:0] chan_cpu_drlin;
   wire [255:0] chan_cpu_data;
   wire [15:0] unexpected_trfr;
   wire [15:0] busy;
   wire [15:0] chan_interrupt;
   
   reg        vdmc_cpu_drlin;
   reg [15:0] vdmc_cpu_data;
   
   reg interrupt;
   reg [3:0] intr_chan;
   reg error;
   reg [3:0] error_chan;
   reg vdmc_pilVD;
   reg vdmc_pil00;

   generate
      for (i=0; i<NDMC; i=i+1) begin
         h316_vdmc_chn
             #(.CHAN(i))
         chan
             (// Outputs
              .chan_cpu_drlin     (chan_cpu_drlin[i]),
              .chan_cpu_data      (chan_cpu_data[16*i +: 16]),
              .DILxx_p            (DILxx_p[i+1]),
              .unexpected_trfr    (unexpected_trfr[i]),
              .busy               (busy[i]),
              .chan_interrupt     (chan_interrupt[i]),
              // Inputs
              .addressed          (addressed[i]),
              .reg_addressed      (reg_addressed[i]),
              .function_code      (function_code),
              .DALxx_n            (DALxx_n[i+1]),
              .reg_DALxx          (reg_DALxx[i]),
              .clk                (clk),
              .reset_N            (reset_N),
              /*AUTOINST*/
              // Inputs
              .reg_fcode         (reg_fcode[3:0]),
              .reg_OTBxx         (reg_OTBxx[15:0]),
              .reg_RRLIN         (reg_RRLIN),
              .reg_OCPLS         (reg_OCPLS),
              .reg_ERLXX         (reg_ERLXX),
              .clear_error       (clear_error));
      end // for (i=0; i<NDMC; i=i+1)

      for (i=NDMC; i<16; i=i+1) begin
         assign chan_cpu_drlin[i] = 1'b0;
         assign chan_cpu_data[16*i +: 16] = 16'd0;
         assign DILxx_p[i] = 1'b0;
         assign unexpected_trfr[i] = 1'b0;
         assign busy[i] = 1'b0;
         assign chan_interrupt[i] = 1'b0;
      end
   endgenerate
   
   assign function_code  = (~ADBxx_n[ 7:10]);
   assign device_address = (~ADBxx_n[11:16]);

   assign ctrl_addressed = (device_address == CTRL_ADDR);
   assign chan_addressed = (device_address == CHAN_ADDR);

   // Drive device-ready line and input bus
   always @(/*AUTOSENSE*/busy or channel or ctrl_addressed or error
            or error_chan or function_code or interrupt or intr_chan) begin
      
      ctrl_cpu_drlin =  1'b0;
      ctrl_cpu_data  = 16'd0;

      if (ctrl_addressed) begin
         case (function_code)

           OTA_CHAN_SLCT: ctrl_cpu_drlin = 1'b1;
           SKS_NOT_BUSY:  ctrl_cpu_drlin = ~(|busy);
           SKS_NTRFR_ERR: ctrl_cpu_drlin = ~error;
           SKS_NOT_INTR:  ctrl_cpu_drlin = ~interrupt;

           INA_CHAN_SLCT: begin ctrl_cpu_drlin = 1'b1; ctrl_cpu_data = {12'd0,    channel}; end
           INA_INTR_CHAN: begin ctrl_cpu_drlin = 1'b1; ctrl_cpu_data = {12'd0,  intr_chan}; end
           INA_ERR_CHAN:  begin ctrl_cpu_drlin = 1'b1; ctrl_cpu_data = {12'd0, error_chan}; end

           default: begin /* Do Nothing */ end
         endcase // case (function_code)
      end // if (ctrl_addressed)
   end // always @ (...

   // Convert from asynchronous CPU bus to modern clocked
   // synchronous logic.

   assign strobe_N = OCPLS_n & RRLIN_n;

   always @(posedge strobe_N or negedge OCPLS_n) begin
      if (!OCPLS_n) begin
         lat_OCPLS_n <= 1'b0;
      end else begin
         lat_OCPLS_n <= 1'b1;
      end
   end
   
   always @(posedge strobe_N or negedge RRLIN_n) begin
      if (!RRLIN_n) begin
         lat_RRLIN_n <= 1'b0;
      end else begin
         lat_RRLIN_n <= 1'b1;
      end
   end

   always @(posedge strobe_N or negedge ERLXX_n) begin
      if (!ERLXX_n) begin
         lat_ERLXX_n <= 1'b0;
      end else begin
         lat_ERLXX_n <= 1'b1;
      end
   end

   // Interrupt mask
   always @(posedge SMK01_n or negedge MSTCL_n) begin
      if (!MSTCL_n) begin
         intr_mask <= 1'b0;
      end else begin
         intr_mask <= OTBxx_p[SMK_BIT];
      end
   end

   always @(posedge strobe_N or negedge MSTCL_n) begin
      if (!MSTCL_n) begin
         reg_OCPLS     <= 1'b0;
         reg_RRLIN     <= 1'b0;
         reg_ctrl_addr <= 1'b0;
         reg_chan_addr <= 1'b0;
         reg_fcode     <= 4'd0;
         reg_OTBxx     <= 16'd0;
         reg_ERLXX     <= 1'b0;
         tmp_DALxx     <= 16'd0;
         bus_update    <= 1'b0;
      end else if (strobe_N) begin // if (!MSTCL_n)
         reg_OCPLS     <= ~lat_OCPLS_n;
         reg_RRLIN     <= ~lat_RRLIN_n;
         reg_ctrl_addr <=  ctrl_addressed;
         reg_chan_addr <=  chan_addressed;
         reg_fcode     <=  function_code;
         reg_OTBxx     <=  OTBxx_p;
         reg_ERLXX     <= ~lat_ERLXX_n;
         tmp_DALxx     <= ~DALxx_n;
         bus_update    <= ~bus_update;
      end // if (strobe_N)
   end // always @ (posedge strobe_N or negedge MSTCL_n)

   assign strobe = bus_update ^ clk_update;

   always @(/*AUTOSENSE*/strobe or tmp_DALxx) begin : p_DAL
      integer j;
      for (j=0; j<16; j=j+1) begin
         reg_DALxx[j] = tmp_DALxx[j+1] & strobe;
      end
   end
   
   always @(/*AUTOSENSE*/channel or clk_update or reg_OCPLS
            or reg_OTBxx or reg_RRLIN or reg_ctrl_addr or reg_fcode
            or strobe) begin

      next_clk_update = clk_update;
      next_channel    = channel;
      clear_error     = 1'b0;
      
      if (strobe) begin
         next_clk_update = ~clk_update;

         if (reg_ctrl_addr) begin
            if (reg_OCPLS) begin // OCP
               if (reg_fcode == OCP_CLEAR_ERR) begin
                  clear_error = 1'b1;
               end
               
            end else if (reg_RRLIN) begin
               if (reg_fcode == OTA_CHAN_SLCT) begin
                  next_channel = reg_OTBxx[3:0];
               end
            end
         end // if (reg_ctrl_addr)
      end // if (strobe)
   end // always @ (...

   always @(/*AUTOSENSE*/chan_addressed or channel) begin
      addressed = 16'd0;
      if (chan_addressed) begin
         addressed[channel] = 1'b1;
      end
   end

   always @(/*AUTOSENSE*/channel or reg_chan_addr or strobe) begin
      reg_addressed = 16'd0;
      if (reg_chan_addr && strobe) begin
         reg_addressed[channel] = 1'b1;
      end
   end
   
   always @(posedge clk) begin
      if (!reset_N) begin
         clk_update <= 1'b0;
         channel    <= 4'd0;
      end else begin
         clk_update <= next_clk_update;
         channel    <= next_channel;
      end
   end

   always @(/*AUTOSENSE*/chan_cpu_data or chan_cpu_drlin
            or ctrl_cpu_data or ctrl_cpu_drlin) begin : chns_d
      integer i;

      vdmc_cpu_drlin = ctrl_cpu_drlin;
      vdmc_cpu_data  = ctrl_cpu_data;
      
      for (i=0; i<NDMC; i=i+1) begin
         vdmc_cpu_drlin = vdmc_cpu_drlin | chan_cpu_drlin[i];
         vdmc_cpu_data  = vdmc_cpu_data  | chan_cpu_data[16*i +: 16];
      end
   end // block: chns_d

   /*
    * Determine the highest priority interrupting channel
    */
   always @(/*AUTOSENSE*/chan_interrupt or channel or intr_mask) begin : intr_p
      integer i;

      interrupt = 1'b0;
      intr_chan = 4'd15;
      
      vdmc_pilVD = 1'b0;
      vdmc_pil00 = 1'b0;
      
      for (i=NDMC-1; i>=0; i=i-1) begin : pi
         if (chan_interrupt[i]) begin
            interrupt = 1'b1;
            intr_chan = i;
         end
      end
      
      if (interrupt & (intr_chan <= channel))
        vdmc_pilVD = 1'b1;

      if (vdmc_pilVD && intr_mask)
        vdmc_pil00 = 1'b1;
      
   end // block: intr_p

   /*
    * Determine the highest priority channel with unexpected transfer error
    */
   always @(/*AUTOSENSE*/unexpected_trfr) begin : err_p
      integer i;

      error = 1'b0;
      error_chan = 4'd15;
      
      for (i=NDMC-1; i>=0; i=i-1) begin : pi
         if (unexpected_trfr[i]) begin
            error = 1'b1;
            error_chan = i;
         end
      end
      
   end // block: err_p

   generate for (i=1; i<=16; i=i+1) begin
      nand (strong0, pull1) #16 u_INBxx_n (INBxx_n[i], vdmc_cpu_data[16-i]);      
   end endgenerate
   
   nand (strong0, pull1) #16 u_DRLIN_n (DRLIN_n, vdmc_cpu_drlin);      
   nand (strong0, pull1) #16 u_PIL00_n (PIL00_n, vdmc_pil00);      
   nand (strong0, pull1) #16 u_PILVD_n (PILVD_n, vdmc_pilVD);      
   
endmodule // h316_vdmc
