/***********************************************************************
 *
 * h316 - Simulation ASR monitor/driver
 * 
 * Copyright (C) 2017 Adrian Wise
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See http://www.series16.adrianwise.co.uk
 *   mailto:apw@adrianwise.co.uk
 * 
 ***********************************************************************
 * 
 * This testbench component models an ASR connected to the current-loop
 * interface provided by the gate-level H316. Characters received from
 * the CPU are assembled in a line-buffer and printed (using $display)
 * each time a line-feed character is received. Such messages are
 * prefixed with "ASR:".
 * 
 * Because many H316 programs tended to print carriage-return, line-feed
 * at the *start* of messages, leaving the carriage at the right of
 * what was printed, in many cases it happens that a message will not
 * be seen until long after it was printed by the CPU, and this can be
 * confusing for debugging. To work-around this issue, a timer is
 * implemented and if the ASR output goes idle for a period after
 * printing characters, the line is printed and the line-feed that would
 * be printed when it is eventually received is suppressed.
 * 
 * A task 'flush' is provided, allowing the testbench to print any
 * characters remaining in the line buffer (for example if the CPU
 * halts).
 * 
 * A task 'send' is provided to allow the testbench to send characters
 * to the CPU over the ASR interface. The task 'wait-sent' may be used
 * to wait until the previous character has been sent (for example, if
 * multiple testbench processes drive the ASR interface).
 * 
 **********************************************************************/

module sim_asr
  #( parameter BAUD = 153600 )
   ( inout TYSIG_n,
     inout MCSET_p);

   localparam TSPS = 1000000000; // Timesteps per second (assume 1ns timescale)
   localparam HALF_BIT = TSPS / (BAUD * 2);
   localparam BIT_TIME = (HALF_BIT * 2);
   localparam ASR_IDLE_LIMIT = 1000;

   reg busy = 1'b0;
   reg send_busy = 1'b0;
   reg TYSIG_p = 1'b0;
   
   event start;
   event sent;

   reg [256*8-1:1] linebuf;
   
   reg clk = 1'b0;
   reg c_flag = 0;
   integer asr_ctr = 0;
   reg lf_flag = 0;
   
   always @(posedge MCSET_p) begin
      clk <= ~clk;
   end

  always @(negedge TYSIG_n) begin
      if (!busy) begin
         -> start;
      end
   end

   initial begin : p_mon
      integer i;
      reg [7:0] data;
      
      busy = 1'b0;
      linebuf = 0;
      
      forever begin
         @(start);
         busy = 1'b1;

         #HALF_BIT; // Get to middle of start bit
         #BIT_TIME; // Get to middle of first data bit

         for (i = 0; i < 8; i=i+1) begin
            data[i] = TYSIG_n;
            #BIT_TIME;
         end
         // Leave loop at moddle of first stop bit
         
         #BIT_TIME; // Get to middle of second stop bit

         // Go not-busy here expecting TYSIG_n to remain
         // passive for the remaining half bit-time, but
         // don't want to race between not-busy and
         // falling edge that starts the next character.

         c_flag = 1'b1;
         if (data[6:0] > 7'h1f)
           linebuf = linebuf << 8 | data[6:0];
            
         if (data[6:0] == 7'h0a) begin
            if (linebuf == 0)
              if (lf_flag) begin
                 // Line already printed
                 lf_flag = 0;
              end else begin
                 $display("ASR:");
              end
            else
              $display("ASR: %0s", linebuf);
            linebuf = 0;
         end
         
         busy = 1'b0;
      end
   end // block: p_mon

   always @(posedge clk) begin
      if (c_flag) begin
         asr_ctr <= 0;
         c_flag <= 0;
      end else begin
         if (asr_ctr < ASR_IDLE_LIMIT) begin
            asr_ctr <= asr_ctr + 1;
         end else if (asr_ctr == ASR_IDLE_LIMIT) begin
            asr_ctr <= asr_ctr + 1;
            if (linebuf != 0) begin
               $display("ASR: %0s", linebuf);
               linebuf <= 0;
               lf_flag <= 1;
            end
         end
      end
   end
   
   task flush;
      if (linebuf != 0) begin
        $display("ASR: %0s", linebuf);
      end
   endtask // flush

   initial begin
      send_busy = 1'b0;
   end

   task wait_sent;
      begin
         while (send_busy) begin
            @(sent);
         end
      end
   endtask // wait_sent

   task send(input [7:0] data);
      begin : p_drv
         integer i;
         
         wait_sent;

         send_busy = 1'b1;

         TYSIG_p = 1'b1;
         #BIT_TIME;

         for (i = 0; i < 8; i=i+1) begin
            TYSIG_p = ~data[i];
            #BIT_TIME;
         end
         
         TYSIG_p = 1'b0;
         #BIT_TIME;
         #BIT_TIME;

         send_busy = 1'b0;
         ->sent;
      end
   endtask // send

   nand (strong0, pull1) #16 u_TYSIG_n (TYSIG_n, TYSIG_p);
   
endmodule // sim_asr
